from datetime import date
from json import dumps

from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import Q
from django.forms.models import inlineformset_factory
from django.forms.widgets import Media
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from sentry_sdk import capture_message

from ..models.agegroup import AgeGroup
from ..models.citizenship import Citizenship
from ..models.courses import Course, CourseRegistration, CourseRegistrationPeriod
from ..models.department import Department
from ..models.events import Event, EventRegistration
from ..models.fields import DAY_OF_WEEK
from ..models.leprikonsite import LeprikonSite
from ..models.orderables import Orderable, OrderableRegistration
from ..models.place import Place
from ..models.roles import BillingInfo, Leader, Parent, Participant
from ..models.school import School
from ..models.schoolyear import SchoolYearPeriod
from ..models.subjects import (
    Subject,
    SubjectGroup,
    SubjectRegistration,
    SubjectRegistrationBillingInfo,
    SubjectRegistrationGroup,
    SubjectRegistrationGroupMember,
    SubjectRegistrationParticipant,
    SubjectType,
    SubjectVariant,
)
from ..models.targetgroup import TargetGroup
from ..utils import get_age, get_birth_date, get_gender
from .fields import AgreementBooleanField
from .form import FormMixin
from .widgets import CheckboxSelectMultipleBootstrap, RadioSelectBootstrap

User = get_user_model()


class RegistrationsFilterForm(FormMixin, forms.Form):
    q = forms.CharField(label=_("Search term"), required=False)
    not_paid = forms.BooleanField(label=_("Not paid"), required=False)


class SubjectFilterForm(FormMixin, forms.Form):
    q = forms.CharField(label=_("Search term"), required=False)
    course_types = forms.ModelMultipleChoiceField(queryset=None, label=_("Course type"), required=False)
    event_types = forms.ModelMultipleChoiceField(queryset=None, label=_("Event type"), required=False)
    orderable_types = forms.ModelMultipleChoiceField(queryset=None, label=_("Orderable event type"), required=False)
    departments = forms.ModelMultipleChoiceField(queryset=None, label=_("Department"), required=False)
    groups = forms.ModelMultipleChoiceField(queryset=None, label=_("Group"), required=False)
    leaders = forms.ModelMultipleChoiceField(queryset=None, label=_("Leader"), required=False)
    places = forms.ModelMultipleChoiceField(queryset=None, label=_("Place"), required=False)
    age_groups = forms.ModelMultipleChoiceField(queryset=None, label=_("Age group"), required=False)
    target_groups = forms.ModelMultipleChoiceField(queryset=None, label=_("Target group"), required=False)
    days_of_week = forms.MultipleChoiceField(
        label=_("Day of week"), choices=tuple(sorted(DAY_OF_WEEK.items())), required=False
    )
    past = forms.BooleanField(label=_("Show past"), required=False)
    reg_active = forms.BooleanField(label=_("Available for registration"), required=False)
    invisible = forms.BooleanField(label=_("Show invisible"), required=False)

    _models = {
        SubjectType.COURSE: Course,
        SubjectType.EVENT: Event,
        SubjectType.ORDERABLE: Orderable,
    }

    def __init__(self, subject_type_type, subject_types, school_year, is_staff, data, **kwargs):
        super().__init__(data=data, **kwargs)
        self.subject_type_type = subject_type_type

        # pre filter subjects by initial params
        qs = self._models[subject_type_type].objects
        if subject_type_type != SubjectType.EVENT or data.get("past"):
            qs = qs.filter(school_year=school_year)
        if len(subject_types) == 1:
            qs = qs.filter(subject_type=subject_types[0])
        else:
            qs = qs.filter(subject_type__in=subject_types)
        if not is_staff or "invisible" not in data:
            qs = qs.filter(public=True)
        self.qs = qs

        subject_ids = tuple(qs.order_by("id").values_list("id", flat=True).distinct())
        if len(subject_types) == 1:
            del self.fields["course_types"]
            del self.fields["event_types"]
            del self.fields["orderable_types"]
        elif subject_type_type == SubjectType.COURSE:
            del self.fields["event_types"]
            del self.fields["orderable_types"]
            self.fields["course_types"].queryset = SubjectType.objects.filter(id__in=(st.id for st in subject_types))
        elif subject_type_type == SubjectType.EVENT:
            del self.fields["course_types"]
            del self.fields["orderable_types"]
            self.fields["event_types"].queryset = SubjectType.objects.filter(id__in=(st.id for st in subject_types))
        elif subject_type_type == SubjectType.ORDERABLE:
            del self.fields["course_types"]
            del self.fields["event_types"]
            self.fields["orderable_types"].queryset = SubjectType.objects.filter(id__in=(st.id for st in subject_types))

        self.fields["departments"].queryset = Department.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields["departments"].queryset.count() == 0:
            del self.fields["departments"]

        self.fields["groups"].queryset = SubjectGroup.objects.filter(
            subject_types__in=subject_types, subjects__id__in=subject_ids
        ).distinct()
        if self.fields["groups"].queryset.count() == 0:
            del self.fields["groups"]

        self.fields["leaders"].queryset = (
            Leader.objects.filter(subjects__id__in=subject_ids)
            .distinct()
            .order_by("user__first_name", "user__last_name")
        )
        if self.fields["leaders"].queryset.count() == 0:
            del self.fields["leaders"]

        self.fields["places"].queryset = Place.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields["places"].queryset.count() == 0:
            del self.fields["places"]

        self.fields["age_groups"].queryset = AgeGroup.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields["age_groups"].queryset.count() == 0:
            del self.fields["age_groups"]

        self.fields["target_groups"].queryset = TargetGroup.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields["target_groups"].queryset.count() == 0:
            del self.fields["target_groups"]

        if subject_type_type != SubjectType.COURSE:
            del self.fields["days_of_week"]
        if subject_type_type != SubjectType.EVENT:
            del self.fields["past"]
        if not is_staff:
            del self.fields["invisible"]

        for f in self.fields:
            self.fields[f].help_text = None

    def get_queryset(self):
        if not self.is_valid():
            return self.qs
        qs = self.qs
        for word in self.cleaned_data["q"].split():
            qs = qs.filter(Q(name__icontains=word) | Q(description__icontains=word))
        if self.cleaned_data.get("course_types"):
            qs = qs.filter(subject_type__in=self.cleaned_data["course_types"])
        elif self.cleaned_data.get("event_types"):
            qs = qs.filter(subject_type__in=self.cleaned_data["event_types"])
        elif self.cleaned_data.get("orderable_types"):
            qs = qs.filter(subject_type__in=self.cleaned_data["orderable_types"])
        if self.cleaned_data.get("departments"):
            qs = qs.filter(department__in=self.cleaned_data["departments"])
        if self.cleaned_data.get("groups"):
            qs = qs.filter(groups__in=self.cleaned_data["groups"])
        if self.cleaned_data.get("places"):
            qs = qs.filter(place__in=self.cleaned_data["places"])
        if self.cleaned_data.get("leaders"):
            qs = qs.filter(leaders__in=self.cleaned_data["leaders"])
        if self.cleaned_data.get("age_groups"):
            qs = qs.filter(age_groups__in=self.cleaned_data["age_groups"])
        if self.cleaned_data.get("target_groups"):
            qs = qs.filter(target_groups__in=self.cleaned_data["target_groups"])
        if self.cleaned_data.get("days_of_week"):
            qs = qs.filter(times__day_of_week__in=self.cleaned_data["days_of_week"])
        if self.subject_type_type == SubjectType.EVENT:
            if self.cleaned_data["past"]:
                qs = qs.filter(end_date__lte=now()).order_by("-start_date", "-start_time")
            else:
                qs = qs.filter(end_date__gte=now())
        if self.cleaned_data["reg_active"]:
            qs = qs.filter(variants__reg_from__lte=now()).exclude(variants__reg_to__lte=now())
        return qs.distinct()


class SubjectForm(FormMixin, forms.ModelForm):
    class Meta:
        model = Subject
        fields = ["description"]


class SubjectAdminForm(forms.ModelForm):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        instance = kwargs.get("instance")
        leprikon_site = LeprikonSite.objects.get_current()

        # limit choices of groups
        self.fields["groups"].widget.choices.queryset = self.subject_type.groups.all()

        # limit choices of leaders
        leaders_choices = self.fields["leaders"].widget.choices
        leaders_choices.queryset = leaders_choices.queryset.filter(school_years=self.school_year)

        # limit choices of questions
        if instance:
            questions_choices = self.fields["questions"].widget.choices
            questions_choices.queryset = questions_choices.queryset.exclude(
                id__in=instance.subject_type.questions.values("id")
            )

        # limit choices of registration agreements
        registration_agreements_choices = self.fields["registration_agreements"].widget.choices
        registration_agreements_choices.queryset = registration_agreements_choices.queryset.exclude(
            id__in=leprikon_site.registration_agreements.values("id")
        )
        if instance:
            registration_agreements_choices.queryset = registration_agreements_choices.queryset.exclude(
                id__in=instance.subject_type.registration_agreements.values("id")
            )


class SubjectVariantForm(forms.ModelForm):
    subject: Subject

    class Meta:
        model = SubjectVariant
        exclude = ("subject", "order")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.subject = self.subject
        if self.subject.subject_type.subject_type == SubjectType.COURSE:
            school_year_divisions = self.fields["school_year_division"].widget.choices.queryset.filter(
                school_year=self.subject.school_year
            )
            self.fields["school_year_division"].widget.choices.queryset = school_year_divisions
        if self.subject.registration_type_participants:
            self.fields["age_groups"].widget.choices.queryset = self.subject.age_groups.all()
        if self.subject.registration_type_groups:
            self.fields["target_groups"].widget.choices.queryset = self.subject.target_groups.all()
        if SubjectVariant.objects.filter(subject=self.subject).exclude(id=self.instance.id).count():
            self.fields["name"].required = True


class SchoolMixin:
    def _add_error_required(self, field_name):
        self.add_error(
            field_name,
            ValidationError(
                self.fields[field_name].error_messages["required"],
                code="required",
            ),
        )

    def clean(self):
        # check required school and class
        x_group = self.cleaned_data.get(self.x_group)
        if x_group and x_group.require_school:
            # require school
            school = self.cleaned_data.get("school")
            if school:
                if school == "other":
                    self.cleaned_data["school"] = None
                    # require other school
                    if not self.cleaned_data.get("school_other"):
                        self._add_error_required("school_other")
                else:
                    self.cleaned_data["school"] = School.objects.get(
                        id=int(self.cleaned_data["school"]),
                    )
            else:
                self._add_error_required("school")
            # require school class
            if not self.cleaned_data.get("school_class"):
                self._add_error_required("school_class")
        else:
            # delete school and class
            self.cleaned_data["school"] = None
            self.cleaned_data["school_other"] = ""
            self.cleaned_data["school_class"] = ""
        return self.cleaned_data


class RegistrationParticipantFormMixin:
    def clean_birth_num(self):
        if self.cleaned_data["birth_num"]:
            qs = SubjectRegistrationParticipant.objects.filter(
                registration__subject_id=self.subject.id,
                registration__canceled=None,
                birth_num=self.cleaned_data["birth_num"],
            )
            if self.instance.pk:
                qs = qs.exclude(pk=self.instance.pk)
            if qs.exists():
                raise ValidationError(_("Participant with this birth number has already been registered."))
        return self.cleaned_data["birth_num"]

    def _setup_required_fields(self):
        self.all_citizenships = list(Citizenship.objects.all())

        # set default values
        created_date = date.today()
        age = None
        citizenship = self.all_citizenships[0]

        # try to use values from existing object (update)
        try:
            created_date = self.obj.created.date()
            age = get_age(self.instance.birth_date, created_date)
            citizenship = self.instance.citizenship
        except AttributeError:
            pass

        prefix = self.prefix + "-"
        data = {key[len(prefix) :]: value for key, value in self.data.items() if key.startswith(prefix)}

        if data:
            # try to use citizenship from request data (post)
            try:
                citizenship = self.fields["citizenship"].clean(data.get("citizenship"))
            except ValidationError:
                pass

        # set required field based on citizenship
        if citizenship.require_birth_num:
            self.hide_birth_num = False
            self.fields["birth_num"].required = True
            self.fields["birth_date"].required = False
            self.fields["gender"].required = False
        else:
            self.hide_birth_num = True
            self.fields["birth_num"].required = False
            self.fields["birth_date"].required = True
            self.fields["gender"].required = True

        if data:
            # try to use age from request data (post)
            try:
                if citizenship.require_birth_num:
                    age = get_age(
                        get_birth_date(self.fields["birth_num"].clean(data.get("birth_num"))),
                        created_date,
                    )
                else:
                    age = get_age(self.fields["birth_date"].clean(data.get("birth_date")), created_date)
            except ValidationError:
                pass

        # set required field based on age
        if age is not None and age < 18:
            self.hide_parents = False
            self.fields["has_parent1"].required = True
        else:
            self.hide_parents = True
            self.fields["phone"].required = True
            self.fields["email"].required = True

        # set required field based on has_parentX
        if data:
            has_parent = ["has_parent1" in data, "has_parent2" in data]
        else:
            has_parent = [self.instance.has_parent1, self.instance.has_parent2] if self.instance else [False, False]
        for n in range(2):
            if has_parent[n]:
                for field in ["first_name", "last_name", "street", "city", "postal_code", "phone", "email"]:
                    self.fields["parent{}_{}".format(n + 1, field)].required = True

    def _set_birth_date_gender(self):
        if self.instance.citizenship.require_birth_num:
            self.instance.birth_date = get_birth_date(self.instance.birth_num)
            self.instance.gender = get_gender(self.instance.birth_num)
        else:
            self.instance.birth_num = None


class RegistrationParticipantForm(FormMixin, RegistrationParticipantFormMixin, SchoolMixin, forms.ModelForm):
    x_group = "age_group"

    school = forms.ChoiceField(
        label=_("School"),
        choices=lambda: (
            [("", "---------")]
            + [(school.id, school) for school in School.objects.all()]
            + [("other", _("other school"))]
        ),
        required=False,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # choices for age group
        self.fields["age_group"].widget.choices.queryset = self.subject.age_groups.all()
        if self.subject.age_groups.count() == 1:
            self.fields["age_group"].initial = self.subject.age_groups.first()

        # initial school
        if School.objects.count() == 0:
            self.fields["school"].initial = "other"

        self._setup_required_fields()

        class ParticipantSelectForm(FormMixin, forms.Form):
            participant = forms.ChoiceField(
                label=_("Choose"),
                choices=[("new", _("new participant"))] + list((p.id, p) for p in self.user_participants),
                initial="new",
                widget=RadioSelectBootstrap(),
            )

        kwargs["prefix"] = self.prefix + "-participant_select"
        self.participant_select_form = ParticipantSelectForm(**kwargs)

        class ParentSelectForm(FormMixin, forms.Form):
            parent = forms.ChoiceField(
                label=_("Choose"),
                choices=[("new", _("new parent"))] + list((p.id, p) for p in self.user_parents),
                initial="new",
                widget=RadioSelectBootstrap(),
            )

        kwargs["prefix"] = self.prefix + "-parent1_select"
        self.parent1_select_form = ParentSelectForm(**kwargs)
        kwargs["prefix"] = self.prefix + "-parent2_select"
        self.parent2_select_form = ParentSelectForm(**kwargs)

        kwargs["prefix"] = self.prefix + "-questions"
        self.questions_form = type(
            str("QuestionsForm"),
            (FormMixin, forms.Form),
            dict((q.slug, q.get_field()) for q in self.subject.all_questions),
        )(**kwargs)

    @cached_property
    def media(self):
        media = Media()
        for form in self.required_forms:
            media = media + form.media
        return media

    @cached_property
    def required_forms(self):
        return [
            super(),
            self.participant_select_form,
            self.questions_form,
            self.parent1_select_form,
            self.parent2_select_form,
        ]

    def is_valid(self):
        # validate all required forms
        results = [form.is_valid() for form in self.required_forms]
        return all(results)

    @cached_property
    def errors(self):
        return {
            (form.prefix + "-" + field_name if form.prefix else field_name): error
            for form in self.required_forms
            for field_name, error in form.errors.items()
        }

    def save(self, commit=True):
        with transaction.atomic():
            return self._save(commit)

    def _save(self, commit):
        # set answers
        self.instance.answers = dumps(self.questions_form.cleaned_data)

        # set birth date and gender
        self._set_birth_date_gender()

        # create
        super().save(commit)

        # save / update participant
        if "participant" not in self.participant_select_form.cleaned_data:
            capture_message(
                "Invalid participant_select_form",
                "warning",
                extras={
                    "participant_select_form_data": self.participant_select_form.data,
                    "participant_select_form_is_valid": self.participant_select_form.is_valid(),
                    "participant_select_form_cleaned_data": self.participant_select_form.cleaned_data,
                },
            )
            self.participant_select_form.cleaned_data["participant"] = "new"
        if self.participant_select_form.cleaned_data["participant"] == "new":
            participant = Participant()
            participant.user = self.instance.registration.user
        else:
            participant = Participant.objects.get(id=self.participant_select_form.cleaned_data["participant"])
        for attr in [
            "first_name",
            "last_name",
            "birth_num",
            "birth_date",
            "gender",
            "age_group",
            "street",
            "city",
            "postal_code",
            "citizenship",
            "phone",
            "email",
            "school",
            "school_other",
            "school_class",
            "health",
        ]:
            setattr(participant, attr, getattr(self.instance, attr))
        participant_answers = participant.get_answers()
        participant_answers.update(self.instance.get_answers())
        participant.answers = dumps(participant_answers)
        participant.save()

        # save / update first parent
        if self.instance.parent1:
            if self.parent1_select_form.cleaned_data["parent"] == "new":
                parent = Parent()
                parent.user = self.instance.registration.user
            else:
                parent = Parent.objects.get(id=self.parent1_select_form.cleaned_data["parent"])
            for attr in ["first_name", "last_name", "street", "city", "postal_code", "phone", "email"]:
                setattr(parent, attr, getattr(self.instance.parent1, attr))
            parent.save()

        # save / update second parent
        if self.instance.parent2:
            if self.parent2_select_form.cleaned_data["parent"] == "new":
                parent = Parent()
                parent.user = self.instance.registration.user
            else:
                parent = Parent.objects.get(id=self.parent2_select_form.cleaned_data["parent"])
            for attr in ["first_name", "last_name", "street", "city", "postal_code", "phone", "email"]:
                setattr(parent, attr, getattr(self.instance.parent2, attr))
            parent.save()
        return self.instance


class RegistrationGroupForm(FormMixin, SchoolMixin, forms.ModelForm):
    x_group = "target_group"

    school = forms.ChoiceField(
        label=_("School"),
        choices=lambda: (
            [("", "---------")]
            + [(school.id, school) for school in School.objects.all()]
            + [("other", _("other school"))]
        ),
        required=False,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # choices for target group
        self.fields["target_group"].widget.choices.queryset = self.subject.target_groups.all()
        if self.subject.target_groups.count() == 1:
            self.fields["target_group"].initial = self.subject.target_groups.first()

        # initial school
        self.fields["school"].initial = "other" if School.objects.count() == 0 else None

        kwargs["prefix"] = self.prefix + "-questions"
        self.questions_form = type(
            str("QuestionsForm"),
            (FormMixin, forms.Form),
            dict((q.slug, q.get_field()) for q in self.subject.all_questions),
        )(**kwargs)

    @cached_property
    def media(self):
        media = Media()
        for form in self.required_forms:
            media = media + form.media
        return media

    @cached_property
    def required_forms(self):
        return [
            super(),
            self.questions_form,
        ]

    def is_valid(self):
        # validate all required forms
        results = [form.is_valid() for form in self.required_forms]
        return all(results)

    @cached_property
    def errors(self):
        return {
            (form.prefix + "-" + field_name if form.prefix else field_name): error
            for form in self.required_forms
            for field_name, error in form.errors.items()
        }

    def save(self, commit=True):
        # set answers
        self.instance.answers = dumps(self.questions_form.cleaned_data)

        # create
        return super().save(commit)


class RegistrationGroupMemberForm(FormMixin, forms.ModelForm):
    pass


class RegistrationBillingInfoForm(FormMixin, forms.ModelForm):
    use_required_attribute = False

    class Meta:
        model = SubjectRegistrationBillingInfo
        exclude = ["registration"]


class RegistrationForm(FormMixin, forms.ModelForm):
    instance: SubjectRegistration

    def __init__(self, subject: Subject, subject_variant: SubjectVariant, user, **kwargs):
        super().__init__(**kwargs)
        self.user = user
        self.instance.subject = subject
        self.instance.subject_variant = subject_variant

        # sub forms
        if subject.registration_type_participants:
            registered_birth_nums = list(
                SubjectRegistrationParticipant.objects.filter(
                    registration_id__in=self.instance.subject.active_registrations.values("id"),
                    birth_num__isnull=False,
                ).values_list("birth_num", flat=True)
            )

            participant_form = type(
                RegistrationParticipantForm.__name__,
                (RegistrationParticipantForm,),
                {
                    "subject": subject,
                    "user_participants": user.leprikon_participants.exclude(
                        birth_num__in=registered_birth_nums,
                    ),
                    "user_parents": list(user.leprikon_parents.all()),
                },
            )

            self.participants_formset = inlineformset_factory(
                SubjectRegistration,
                SubjectRegistrationParticipant,
                form=participant_form,
                fk_name="registration",
                exclude=[],
                extra=subject_variant.max_participants_count,
                min_num=subject_variant.min_participants_count,
                max_num=subject_variant.max_participants_count,
                validate_min=True,
                validate_max=True,
            )(kwargs.get("data"), instance=self.instance)

        elif subject.registration_type_groups:
            group_form = type(RegistrationGroupForm.__name__, (RegistrationGroupForm,), {"subject": subject})

            self.group_formset = inlineformset_factory(
                SubjectRegistration,
                SubjectRegistrationGroup,
                form=group_form,
                fk_name="registration",
                exclude=[],
                extra=1,
                min_num=1,
                max_num=1,
                validate_min=True,
                validate_max=True,
            )(kwargs.get("data"), instance=self.instance)

            self.group_members_formset = inlineformset_factory(
                SubjectRegistration,
                SubjectRegistrationGroupMember,
                form=RegistrationGroupMemberForm,
                fk_name="registration",
                exclude=[],
                extra=subject_variant.max_participants_count,
                min_num=subject_variant.min_participants_count,
                max_num=subject_variant.max_participants_count,
                validate_min=True,
                validate_max=True,
            )(kwargs.get("data"), instance=self.instance)

        del kwargs["instance"]

        self.agreement_forms = []
        for agreement in self.instance.subject.all_registration_agreements:
            kwargs["prefix"] = "agreement_%s" % agreement.id
            form_attributes = {
                "agreement": agreement,
                "options": {},
            }
            for option in agreement.all_options:
                option_id = "option_%s" % option.id
                form_attributes[option_id] = AgreementBooleanField(
                    label=option.option,
                    allow_disagree=not option.required,
                )
                form_attributes["options"][option_id] = option
            AgreementForm = type(str("AgreementForm"), (FormMixin, forms.Form), form_attributes)
            self.agreement_forms.append(AgreementForm(**kwargs))

        self.user_billing_info = list(self.user.leprikon_billing_info.all())

        class BillingInfoSelectForm(FormMixin, forms.Form):
            billing_info = forms.ChoiceField(
                label=_("Choose"),
                choices=[
                    ("none", _("no billing information")),
                ]
                + list((bi.id, bi) for bi in self.user_billing_info)
                + [
                    ("new", _("new billing information")),
                ],
                initial="none",
                widget=RadioSelectBootstrap(),
            )

        kwargs["prefix"] = "billing_info_select"
        self.billing_info_select_form = BillingInfoSelectForm(**kwargs)

        kwargs["prefix"] = "billing_info"
        self.billing_info_form = RegistrationBillingInfoForm(**kwargs)

    @cached_property
    def media(self):
        media = Media()
        for form in self.all_forms:
            media = media + form.media
        return media

    @cached_property
    def required_forms(self):
        return [
            super(),
            self.billing_info_select_form,
        ] + self.agreement_forms

    @cached_property
    def all_forms(self):
        all_forms = self.required_forms
        if self.instance.subject.registration_type_participants:
            all_forms += self.participants_formset.forms
        elif self.instance.subject.registration_type_groups:
            all_forms += self.group_formset.forms + self.group_members_formset.forms
        if (
            not self.billing_info_select_form.is_valid()
            or self.billing_info_select_form.cleaned_data["billing_info"] != "none"
        ):
            all_forms.append(self.billing_info_form)
        return all_forms

    def is_valid(self):
        # validate all required forms
        results = [form.is_valid() for form in self.required_forms]
        if self.instance.subject.registration_type_participants:
            results.append(self.participants_formset.is_valid())
        elif self.instance.subject.registration_type_groups:
            results += [
                self.group_formset.is_valid(),
                self.group_members_formset.is_valid(),
            ]
        if (
            not self.billing_info_select_form.is_valid()
            or self.billing_info_select_form.cleaned_data["billing_info"] != "none"
        ):
            results.append(self.billing_info_form.is_valid())
        return all(results)

    @cached_property
    def errors(self):
        return {
            (form.prefix + field_name if form.prefix else field_name): error
            for form in self.all_forms
            for field_name, error in form.errors.items()
        }

    @transaction.atomic
    def save(self, commit=True):
        # set user
        self.instance.created_by = self.user
        self.instance.user = self.user

        if self.instance.subject.registration_type_participants:
            participants_count = self.participants_formset.total_form_count()
        elif self.instance.subject.registration_type_groups:
            participants_count = self.group_members_formset.total_form_count()

        # set price
        self.instance.price = self.instance.subject_variant.get_price(participants_count)

        # create
        super().save(True)

        if self.instance.subject.registration_type_participants:
            # save participants
            self.participants_formset.user = self.instance.user
            self.participants_formset.save()

        elif self.instance.subject.registration_type_groups:
            # save group
            self.group_formset.save()

            # save group members
            self.group_members_formset.save()

        # save additional questions
        self.instance.questions.set(self.instance.subject.all_questions)

        # save legal agreements
        self.instance.agreements.set(self.instance.subject.all_registration_agreements)

        # save agreement options
        for agreement_form in self.agreement_forms:
            for option_id, checked in agreement_form.cleaned_data.items():
                if checked:
                    self.instance.agreement_options.add(agreement_form.options[option_id])

        if self.billing_info_select_form.cleaned_data["billing_info"] != "none":
            # save registration billing info
            registration_billing_info = self.billing_info_form.save(commit=False)
            registration_billing_info.registration = self.instance
            registration_billing_info.save()

            # save / update user billing info
            if self.billing_info_select_form.cleaned_data["billing_info"] == "new":
                billing_info = BillingInfo()
                billing_info.user = self.instance.user
            else:
                billing_info = BillingInfo.objects.get(id=self.billing_info_select_form.cleaned_data["billing_info"])
            for attr in [
                "name",
                "street",
                "city",
                "postal_code",
                "company_num",
                "vat_number",
                "contact_person",
                "phone",
                "email",
                "employee",
            ]:
                setattr(billing_info, attr, getattr(registration_billing_info, attr))
            billing_info.save()

        self.instance.generate_variable_symbol_and_slug()
        self.instance.send_mail()
        return self.instance


class CourseRegistrationForm(RegistrationForm):
    periods = forms.ModelMultipleChoiceField(
        queryset=SchoolYearPeriod.objects.all(),
        widget=CheckboxSelectMultipleBootstrap(),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.subject_variant.school_year_division:
            self.available_periods = self.instance.subject_variant.school_year_division.periods.exclude(
                end__lt=date.today(),
            )
        else:
            self.available_periods = []

        if self.instance.subject_variant.school_year_division and self.instance.subject_variant.allow_period_selection:
            self.fields["periods"].widget.choices.queryset = self.available_periods
        else:
            del self.fields["periods"]

    class Meta:
        model = CourseRegistration
        exclude = (
            "subject",
            "subject_variant",
            "user",
            "cancelation_requested",
            "cancelation_requested_by",
            "canceled",
            "canceled_by",
        )

    @transaction.atomic
    def save(self, commit=True):
        super().save()
        CourseRegistrationPeriod.objects.bulk_create(
            CourseRegistrationPeriod(
                registration=self.instance,
                period=period,
            )
            for period in self.cleaned_data.get("periods", self.available_periods)
        )
        return self.instance


class EventRegistrationForm(RegistrationForm):
    class Meta:
        model = EventRegistration
        exclude = (
            "subject",
            "subject_variant",
            "user",
            "cancelation_requested",
            "cancelation_requested_by",
            "canceled",
            "canceled_by",
        )


class OrderableRegistrationForm(RegistrationForm):
    class Meta:
        model = OrderableRegistration
        exclude = (
            "subject",
            "subject_variant",
            "user",
            "cancelation_requested",
            "cancelation_requested_by",
            "canceled",
            "canceled_by",
        )


class RegistrationAdminForm(forms.ModelForm):
    instance: SubjectRegistration
    subject: Subject
    subject_variant: SubjectVariant

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)

        if "subject_variant" in self.fields:
            self.fields["subject_variant"].widget.choices.queryset = self.subject.variants.all()

        # choices for agreement options
        if "agreement_options" in self.fields:
            instance: SubjectRegistration = kwargs.get("instance")
            self.fields["agreement_options"].widget.choices = tuple(
                (agreement.name, tuple((option.id, option.name) for option in agreement.all_options))
                for agreement in (instance.all_agreements if instance else self.subject.all_registration_agreements)
            )


class SchoolAdminMixin:
    def _add_error_required(self, field_name):
        self.add_error(
            field_name,
            ValidationError(
                self.fields[field_name].error_messages["required"],
                code="required",
            ),
        )

    def clean(self):
        # check required school and class
        x_group = self.cleaned_data.get(self.x_group)
        if x_group and x_group.require_school:
            # require school
            if self.cleaned_data.get("school"):
                self.cleaned_data["school_other"] = ""
            elif not self.cleaned_data.get("school_other"):
                self._add_error_required("school_other")
            # require school class
            if not self.cleaned_data.get("school_class"):
                self._add_error_required("school_class")
        else:
            # delete school and class
            self.cleaned_data["school"] = None
            self.cleaned_data["school_other"] = ""
            self.cleaned_data["school_class"] = ""
        return self.cleaned_data


class RegistrationParticipantAdminForm(RegistrationParticipantFormMixin, SchoolAdminMixin, forms.ModelForm):
    x_group = "age_group"

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        instance = kwargs.get("instance")

        self.fields["age_group"].widget.choices.queryset = self.subject.age_groups.all()
        if self.subject.age_groups.count() == 1:
            self.fields["age_group"].initial = self.subject.age_groups.first()

        questions = self.obj.all_questions if self.obj else self.subject.all_questions
        answers = instance.get_answers() if instance else {}
        for q in questions:
            self.fields["q_" + q.slug].initial = answers.get(q.slug)

        self._setup_required_fields()

    def save(self, commit=True):
        self.instance.answers = dumps(
            {
                q.slug: self.cleaned_data["q_" + q.slug]
                for q in (self.obj.all_questions if self.obj else self.subject.all_questions)
            }
        )
        self._set_birth_date_gender()
        return super().save(commit)


class RegistrationGroupAdminForm(SchoolAdminMixin, forms.ModelForm):
    x_group = "target_group"

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        instance = kwargs.get("instance")

        self.fields["target_group"].widget.choices.queryset = self.subject.target_groups.all()

        questions = self.obj.all_questions if self.obj else self.subject.all_questions
        answers = instance.get_answers() if instance else {}
        for q in questions:
            self.fields["q_" + q.slug].initial = answers.get(q.slug)

    def save(self, commit=True):
        self.instance.answers = dumps(
            {
                q.slug: self.cleaned_data["q_" + q.slug]
                for q in (self.obj.all_questions if self.obj else self.subject.all_questions)
            }
        )
        return super().save(commit)



from collections import namedtuple

from django.urls import reverse_lazy as reverse
from django.utils.translation import gettext_lazy as _

from ...forms.reports.courses import CoursePaymentsForm, CoursePaymentsStatusForm, CourseStatsForm
from ...forms.reports.debtors import DebtorsForm
from ...forms.reports.events import EventPaymentsForm, EventPaymentsStatusForm, EventStatsForm
from ...forms.reports.orderables import OrderablePaymentsForm, OrderablePaymentsStatusForm, OrderableStatsForm
from ..generic import TemplateView


class ReportsListView(TemplateView):
    template_name = "leprikon/reports/index.html"

    Report = namedtuple("Report", ("title", "instructions", "form", "url"))

    def get_context_data(self, *args, **kwargs):
        return super().get_context_data(
            reports=[
                self.Report(
                    title=_("Course payments"),
                    instructions="",
                    form=CoursePaymentsForm(),
                    url=reverse("leprikon:report_course_payments"),
                ),
                self.Report(
                    title=_("Event payments"),
                    instructions="",
                    form=EventPaymentsForm(),
                    url=reverse("leprikon:report_event_payments"),
                ),
                self.Report(
                    title=_("Orderable event payments"),
                    instructions="",
                    form=OrderablePaymentsForm(),
                    url=reverse("leprikon:report_orderable_payments"),
                ),
                self.Report(
                    title=_("Course payments status"),
                    instructions="",
                    form=CoursePaymentsStatusForm(),
                    url=reverse("leprikon:report_course_payments_status"),
                ),
                self.Report(
                    title=_("Event payments status"),
                    instructions="",
                    form=EventPaymentsStatusForm(),
                    url=reverse("leprikon:report_event_payments_status"),
                ),
                self.Report(
                    title=_("Orderable event payments status"),
                    instructions="",
                    form=OrderablePaymentsStatusForm(),
                    url=reverse("leprikon:report_orderable_payments_status"),
                ),
                self.Report(
                    title=_("Course statistics"),
                    instructions="",
                    form=CourseStatsForm(school_year=self.request.school_year),
                    url=reverse("leprikon:report_course_stats"),
                ),
                self.Report(
                    title=_("Event statistics"),
                    instructions="",
                    form=EventStatsForm(school_year=self.request.school_year),
                    url=reverse("leprikon:report_event_stats"),
                ),
                self.Report(
                    title=_("Orderable event statistics"),
                    instructions="",
                    form=OrderableStatsForm(school_year=self.request.school_year),
                    url=reverse("leprikon:report_orderable_stats"),
                ),
                self.Report(
                    title=_("Debtors list"),
                    instructions="",
                    form=DebtorsForm(),
                    url=reverse("leprikon:report_debtors"),
                ),
            ]
        )



from django import forms
from django.contrib import admin
from django.http.response import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.urls import path, reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..forms.journals import JournalAdminForm, JournalEntryAdminForm, JournalLeaderEntryAdminForm
from ..models.journals import Journal, JournalEntry, JournalLeaderEntry, JournalTime
from ..models.subjects import Subject
from ..utils import attributes
from .export import AdminExportMixin
from .filters import JournalListFilter, LeaderListFilter, SchoolYearListFilter, SubjectListFilter, SubjectTypeListFilter


class JournalTimeInlineAdmin(admin.TabularInline):
    model = JournalTime
    extra = 0


@admin.register(Journal)
class JournalAdmin(AdminExportMixin, admin.ModelAdmin):
    filter_horizontal = ("leaders", "participants")
    inlines = (JournalTimeInlineAdmin,)
    list_display = ("subject", "name", "get_times_list", "journal_links")
    list_filter = (
        ("subject__school_year", SchoolYearListFilter),
        ("subject__subject_type", SubjectTypeListFilter),
        ("leaders", LeaderListFilter),
        ("subject", SubjectListFilter),
    )
    raw_id_fields = ("subject",)

    # do not allow to add journals in admin
    def has_add_permission(self, request, obj=None):
        return False

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and len(request.POST) == 3:
            return HttpResponseRedirect(f"{request.path}?subject=" + request.POST.get("subject", ""))
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_inline_instances(self, request, obj=None):
        return [] if hasattr(request, "hide_inlines") else super().get_inline_instances(request, obj)

    def get_form(self, request, obj, **kwargs):
        # get subject
        try:
            # first try request.POST (user may want to change subject)
            request.subject = Subject.objects.get(id=int(request.POST.get("subject")))
        except (Subject.DoesNotExist, TypeError, ValueError):
            if obj:
                # use subject type from object
                request.subject = obj.subject
            else:
                # try to get subject type from request.GET
                try:
                    request.subject = Subject.objects.get(
                        id=int(request.GET.get("subject")),
                    )
                except (Subject.DoesNotExist, TypeError, ValueError):
                    request.subject = None

        if request.subject:
            kwargs["form"] = type(
                "JournalAdminForm",
                (JournalAdminForm,),
                {"subject": request.subject},
            )
        else:
            kwargs["fields"] = ["subject"]
            request.hide_inlines = True

        return super().get_form(request, obj, **kwargs)

    def get_urls(self):
        return [
            path(
                "<int:journal_id>/journal/",
                self.admin_site.admin_view(self.journal),
                name="leprikon_journal_journal",
            ),
            path(
                "<int:journal_id>/journal-pdf/",
                self.admin_site.admin_view(self.journal_pdf),
                name="leprikon_journal_journal_pdf",
            ),
        ] + super().get_urls()

    def journal(self, request, journal_id):
        journal = self.get_object(request, journal_id)

        return render(
            request,
            "leprikon/journal_journal.html",
            {
                "journal": journal,
                "admin": True,
            },
        )

    def journal_pdf(self, request, journal_id):
        journal = self.get_object(request, journal_id)

        # create PDF response object
        response = HttpResponse(content_type="application/pdf")
        response["Content-Disposition"] = 'attachment; filename="{}"'.format(journal.get_pdf_filename("journal_pdf"))

        # write PDF to response
        return journal.write_pdf("journal_pdf", response)

    @attributes(short_description=_("journal"))
    def journal_links(self, obj):
        return mark_safe(
            "<br/>".join(
                (
                    format_html(
                        '<a href="{url}" title="{title}" target="_blank">{journal}</a>',
                        url=reverse("admin:leprikon_journal_journal", args=[obj.id]),
                        title=_("printable journal"),
                        journal=_("journal"),
                    ),
                    format_html(
                        '<a href="{url}" title="{title}" target="_blank">{participants}</a>',
                        url=reverse("admin:leprikon_journal_journal_pdf", args=[obj.id]),
                        title=_("printable list of participants"),
                        participants=_("participants"),
                    ),
                )
            )
        )


@admin.register(JournalLeaderEntry)
class JournalLeaderEntryAdmin(AdminExportMixin, admin.ModelAdmin):
    form = JournalLeaderEntryAdminForm
    list_display = ("timesheet", "date", "start", "end", "duration", "subject")
    list_filter = (
        ("timesheet__leader", LeaderListFilter),
        ("journal_entry__journal__subject", SubjectListFilter),
    )
    ordering = ("-journal_entry__date", "-start")

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        if obj and obj.timesheet.submitted:
            return False
        return super().has_delete_permission(request, obj)

    def get_readonly_fields(self, request, obj=None):
        if obj and obj.timesheet.submitted:
            return ("start", "end")
        return self.readonly_fields


class JournalLeaderEntryInlineAdmin(admin.TabularInline):
    class form(forms.ModelForm):
        class Meta:
            model = JournalLeaderEntry
            fields = []

    model = JournalLeaderEntry
    ordering = ("journal_entry__date", "start")
    readonly_fields = ("date", "start", "end", "edit_link")

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        if obj:
            # obj may be Timesheet or JournalEntry
            # this inline is used in both JournalEntryAdmin and TimesheetAdmin
            try:
                entries = obj.leader_entries
            except AttributeError:
                entries = obj.journal_entries
            if entries.filter(timesheet__submitted=True).exists():
                return False
        return super().has_delete_permission(request, obj)

    @attributes(short_description="")
    def edit_link(self, obj):
        return mark_safe(
            '<a href="{url}" title="{title}" target="_blank">{edit}</a>'.format(
                url=reverse("admin:leprikon_journalleaderentry_change", args=[obj.id]),
                title=_("update entry"),
                edit=_("edit"),
            )
        )


@admin.register(JournalEntry)
class JournalEntryAdmin(AdminExportMixin, admin.ModelAdmin):
    form = JournalEntryAdminForm
    date_hierarchy = "date"
    list_display = ("id", "journal_name", "date", "start", "end", "duration", "agenda_html")
    list_filter = (
        ("journal__subject__school_year", SchoolYearListFilter),
        ("journal__leaders", LeaderListFilter),
        ("journal", JournalListFilter),
    )
    filter_horizontal = ("participants", "participants_instructed")
    inlines = (JournalLeaderEntryInlineAdmin,)
    ordering = ("-date", "-start")
    readonly_fields = (
        "journal_name",
        "date",
    )

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        if obj:
            if obj.affects_submitted_timesheets:
                return False
            else:
                return super().has_delete_permission(request, obj)
        return False

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:

            def delete_selected(model_admin, request, queryset):
                objs = [
                    journal_entry for journal_entry in queryset.all() if not journal_entry.affects_submitted_timesheets
                ]
                return admin.actions.delete_selected(model_admin, request, objs)

            actions["delete_selected"] = (delete_selected, *actions["delete_selected"][1:])
        return actions

    @attributes(admin_order_field="journal", short_description=_("journal"))
    def journal_name(self, obj):
        return str(obj.journal)

    @attributes(admin_order_field="agenda", short_description=_("agenda"))
    def agenda_html(self, obj):
        return mark_safe(obj.agenda)



# Generated by Django 1.10.7 on 2017-09-04 19:32

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('leprikon', '0002_initial_data'),
    ]

    operations = [
        migrations.CreateModel(
            name='Manager',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('school_years', models.ManyToManyField(related_name='managers', to='leprikon.SchoolYear', verbose_name='school years')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, related_name='leprikon_manager', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'ordering': ('user__first_name', 'user__last_name'),
                'verbose_name': 'manager',
                'verbose_name_plural': 'managers',
            },
        ),
        migrations.AddField(
            model_name='courselistplugin',
            name='managers',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by managers.', related_name='_courselistplugin_managers_+', to='leprikon.Manager', verbose_name='managers'),
        ),
        migrations.AddField(
            model_name='eventlistplugin',
            name='managers',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by managers.', related_name='_eventlistplugin_managers_+', to='leprikon.Manager', verbose_name='managers'),
        ),
        migrations.AddField(
            model_name='subject',
            name='managers',
            field=models.ManyToManyField(blank=True, related_name='subjects', to='leprikon.Manager', verbose_name='managers'),
        ),
    ]



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..utils import attributes, currency
from .fields import BankAccountField
from .subjects import SubjectRegistration


class RefundRequest(models.Model):
    requested = models.DateTimeField(_("requested time"), default=timezone.now, editable=False)
    requested_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("requested by"),
    )
    registration: SubjectRegistration = models.OneToOneField(
        SubjectRegistration,
        on_delete=models.PROTECT,
        related_name="refund_request",
        verbose_name=_("registration"),
    )
    bank_account = BankAccountField(_("bank account number"))

    class Meta:
        app_label = "leprikon"
        ordering = ("requested",)
        verbose_name = _("refund request")
        verbose_name_plural = _("refund requests")

    def __str__(self):
        return f"{self.registration}, {currency(self.amount)}"

    def clean(self):
        super().clean()
        if not self.amount:
            raise ValidationError({"registration": [_("Refund may only be requested for overpaid registrations.")]})

    @attributes(short_description=_("amount"))
    @cached_property
    def amount(self):
        return self.registration and self.registration.payment_status.overpaid



from adminsortable2.admin import SortableAdminMixin, SortableInlineAdminMixin
from django.contrib import admin

from ..models.agreements import Agreement, AgreementOption
from .filters import ActiveListFilter


class AgreementOptionInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = AgreementOption
    extra = 0
    min_num = 1


@admin.register(Agreement)
class AgreementAdmin(SortableAdminMixin, admin.ModelAdmin):
    inlines = (AgreementOptionInlineAdmin,)
    list_display = ("name",)
    list_filter = (ActiveListFilter,)

    def has_delete_permission(self, request, obj=None):
        if obj and obj.registrations.exists():
            return False
        else:
            return super().has_delete_permission(request, obj)

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:

            def delete_selected(model_admin, request, queryset):
                queryset = queryset.filter(registrations__id=None)
                return admin.actions.delete_selected(model_admin, request, queryset)

            actions["delete_selected"] = (delete_selected, *actions["delete_selected"][1:])
        return actions



# Generated by Django 1.11.12 on 2018-07-31 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0012_subject_variants'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='participant',
            name='insurance',
        ),
        migrations.RemoveField(
            model_name='subjectregistration',
            name='participant_insurance',
        ),
        migrations.DeleteModel(
            name='Insurance',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-02-15 23:26
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0066_subject_type_slug_unique"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="subject",
            name="chat_group_type",
        ),
        migrations.RemoveField(
            model_name="subjecttype",
            name="chat_group_type",
        ),
    ]



from django import forms
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.utils.translation import gettext_lazy as _

from ..models.courses import CourseDiscount, CourseRegistration, CourseRegistrationPeriod
from ..models.schoolyear import SchoolYearDivision, SchoolYearPeriod
from .subjects import RegistrationAdminForm


class CourseRegistrationAdminForm(RegistrationAdminForm):
    instance: CourseRegistration
    periods = forms.ModelMultipleChoiceField(
        queryset=SchoolYearPeriod.objects.all(),
        widget=FilteredSelectMultiple(
            verbose_name=_("Periods"),
            is_stacked=False,
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.school_year_division = SchoolYearDivision.objects.get(pk=int(self.data["school_year_division"]))
        except (KeyError, TypeError, ValueError, SchoolYearDivision.DoesNotExist):
            if self.instance.subject_variant_id and self.instance.subject_variant.school_year_division_id:
                self.school_year_division = self.instance.subject_variant.school_year_division
            else:
                self.school_year_division = self.subject_variant.school_year_division
        self.available_periods = self.school_year_division.periods.all()
        self.fields["periods"].widget.choices.queryset = self.available_periods
        self.fields["periods"].initial = self.available_periods.filter(
            course_registration_periods__registration=self.instance,
        )

    def _save_m2m(self):
        super()._save_m2m()
        for period in self.cleaned_data["periods"]:
            CourseRegistrationPeriod.objects.get_or_create(
                registration=self.instance,
                period=period,
            )
        self.instance.course_registration_periods.exclude(period__in=self.cleaned_data["periods"]).delete()
        return self.instance


class CourseDiscountAdminForm(forms.ModelForm):
    class Meta:
        model = CourseDiscount
        fields = ["registration", "registration_period", "accounted", "amount", "explanation"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "registration_period" in self.fields:
            try:
                if self.data:
                    registration_id = int(self.data["registration"])
                elif self.initial:
                    registration_id = int(self.initial["registration"])
                else:
                    registration_id = self.instance.registration_id
            except Exception:
                registration_id = None
            self.fields["registration_period"].widget.choices.queryset = self.fields[
                "registration_period"
            ].widget.choices.queryset.filter(
                registration_id=registration_id,
            )



from django import forms
from django.utils.translation import gettext_lazy as _

from ..models.courses import CourseRegistration
from ..models.events import EventRegistration
from ..models.orderables import OrderableRegistration
from ..models.refundrequest import RefundRequest
from ..models.transaction import Transaction
from ..utils import comma_separated, currency, first_upper
from .fields import ReadonlyField
from .form import FormMixin


class RefundRequestBaseForm(FormMixin, forms.ModelForm):
    def __init__(self, registration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.registration = registration
        self.readonly_fields = [
            ReadonlyField(label=first_upper(registration.subject.subject_type.name), value=registration.subject.name)
        ]
        if registration.subject.registration_type_participants:
            if len(registration.all_participants) > 1:
                label = _("Participants")
            else:
                label = _("Participant")
            self.readonly_fields.append(
                ReadonlyField(label=label, value=comma_separated(registration.all_participants))
            )
        elif registration.subject.registration_type_groups:
            self.readonly_fields.append(ReadonlyField(label=_("Contact person"), value=registration.group.full_name))
            if registration.group.name:
                self.readonly_fields.append(ReadonlyField(label=_("Group name"), value=registration.group.name))
        self.readonly_fields.append(
            ReadonlyField(label=_("Overpaid amount"), value=currency(registration.payment_status.overpaid))
        )


class RefundRequestForm(RefundRequestBaseForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.registration = self.registration
        self.instance.requested_by_id = self.registration.user_id

    class Meta:
        model = RefundRequest
        fields = ["bank_account"]


class PaymentTransferForm(RefundRequestBaseForm):
    instance: Transaction

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        valid_target_registration_ids = [
            registration.id
            for Registration in (CourseRegistration, EventRegistration, OrderableRegistration)
            for registration in Registration.objects.filter(user_id=self.registration.user_id)
            if registration.payment_status.amount_due
        ]
        registration_choices = self.fields["target_registration"].widget.choices
        registration_choices.queryset = registration_choices.queryset.filter(id__in=valid_target_registration_ids)
        self.instance.source_registration = self.registration
        self.instance.accounted_by_id = self.registration.user_id
        self.instance.transaction_type = Transaction.TRANSFER

    def clean(self):
        self.cleaned_data = super().clean()
        target_registration = self.cleaned_data.get("target_registration")
        if target_registration:
            self.instance.amount = min(
                self.registration.payment_status.overpaid,
                target_registration.payment_status.amount_due,
            )
        return self.cleaned_data

    class Meta:
        model = Transaction
        fields = ["target_registration"]


class DonationForm(RefundRequestBaseForm):
    instance: Transaction

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.source_registration = self.registration
        self.instance.accounted_by_id = self.registration.user_id
        self.instance.transaction_type = Transaction.DONATION_TRANSFER
        self.instance.amount = self.registration.payment_status.overpaid

    class Meta:
        model = Transaction
        fields = []



from adminsortable2.admin import SortableAdminMixin
from django.contrib import admin

from ..models.agegroup import AgeGroup


@admin.register(AgeGroup)
class AgeGroupAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name", "min_age", "max_age", "require_school")
    list_editable = ("min_age", "max_age", "require_school")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2021-06-01 08:22
from __future__ import unicode_literals

from decimal import Decimal
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import djangocms_text_ckeditor.fields
import leprikon.models.fields
import leprikon.models.subjects
import multiselectfield.db.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("leprikon", "0059_transactions_data"),
    ]

    operations = [
        migrations.RenameField(model_name="SubjectPayment", old_name="received_by", new_name="accounted_by"),
        migrations.RenameField(model_name="SubjectPayment", old_name="payment_type", new_name="transaction_type"),
        migrations.RenameModel(old_name="SubjectPayment", new_name="Transaction"),
        migrations.CreateModel(
            name="Donation",
            fields=[],
            options={
                "verbose_name": "donation",
                "verbose_name_plural": "donations",
                "proxy": True,
                "indexes": [],
            },
            bases=("leprikon.transaction",),
        ),
        migrations.CreateModel(
            name="SubjectPayment",
            fields=[],
            options={
                "verbose_name": "payment",
                "verbose_name_plural": "payments",
                "proxy": True,
                "indexes": [],
            },
            bases=("leprikon.transaction",),
        ),
        migrations.CreateModel(
            name="SubjectReceivedPayment",
            fields=[],
            options={
                "verbose_name": "received payment",
                "verbose_name_plural": "received payments",
                "proxy": True,
                "indexes": [],
            },
            bases=(leprikon.models.subjects.SubjectPaymentMixin, "leprikon.transaction"),
        ),
        migrations.CreateModel(
            name="SubjectReturnedPayment",
            fields=[],
            options={
                "verbose_name": "returned payment",
                "verbose_name_plural": "returned payments",
                "proxy": True,
                "indexes": [],
            },
            bases=(leprikon.models.subjects.SubjectPaymentMixin, "leprikon.transaction"),
        ),
        migrations.AlterModelOptions(
            name="transaction",
            options={"ordering": ("accounted",), "verbose_name": "transaction", "verbose_name_plural": "transactions"},
        ),
        migrations.RemoveField(
            model_name="transaction",
            name="registration",
        ),
        migrations.RemoveField(
            model_name="transaction",
            name="related_payment",
        ),
        migrations.AddField(
            model_name="coursediscount",
            name="accounted_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="accounted by",
            ),
        ),
        migrations.AddField(
            model_name="coursediscount",
            name="last_updated",
            field=models.DateTimeField(editable=False, null=True, verbose_name="last updated time"),
        ),
        migrations.AddField(
            model_name="coursediscount",
            name="last_updated_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="last updated by",
            ),
        ),
        migrations.AddField(
            model_name="eventdiscount",
            name="accounted_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="accounted by",
            ),
        ),
        migrations.AddField(
            model_name="eventdiscount",
            name="last_updated",
            field=models.DateTimeField(editable=False, null=True, verbose_name="last updated time"),
        ),
        migrations.AddField(
            model_name="eventdiscount",
            name="last_updated_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="last updated by",
            ),
        ),
        migrations.AddField(
            model_name="orderablediscount",
            name="accounted_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="accounted by",
            ),
        ),
        migrations.AddField(
            model_name="orderablediscount",
            name="last_updated",
            field=models.DateTimeField(editable=False, null=True, verbose_name="last updated time"),
        ),
        migrations.AddField(
            model_name="orderablediscount",
            name="last_updated_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="last updated by",
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="donation_print_setup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="donation print setup",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="text_payment_returned",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True, default="", verbose_name="text: payment returned"
            ),
        ),
        migrations.AddField(
            model_name="subjecttype",
            name="text_payment_returned",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True,
                default="",
                help_text="<p>Keep empty to use default value:</p><p>Hello,<br/>\nwe have just returned Your payment. Thank You.<br/>\nPlease see the recipe attached.</p>",
                verbose_name="text: payment returned",
            ),
        ),
        migrations.AddField(
            model_name="transaction",
            name="donor",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="leprikon_donations",
                to=settings.AUTH_USER_MODEL,
                verbose_name="donor user",
            ),
        ),
        migrations.AddField(
            model_name="transaction",
            name="last_updated",
            field=models.DateTimeField(editable=False, null=True, verbose_name="last updated time"),
        ),
        migrations.AddField(
            model_name="transaction",
            name="last_updated_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="last updated by",
            ),
        ),
        migrations.AddField(
            model_name="transaction",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="donations",
                to="leprikon.Organization",
                verbose_name="organization",
            ),
        ),
        migrations.AlterField(
            model_name="subjectattachment",
            name="events",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("registration_received", "registration received"),
                    ("registration_approved", "registration approved"),
                    ("registration_refused", "registration refused"),
                    ("registration_payment_request", "payment requested"),
                    ("registration_canceled", "registration canceled"),
                    ("discount_granted", "discount granted"),
                    ("payment_received", "payment received"),
                    ("payment_returned", "payment returned"),
                ],
                default=[],
                help_text="The attachment will be sent with notification on selected events.",
                max_length=166,
                verbose_name="send when",
            ),
        ),
        migrations.AlterField(
            model_name="subjecttypeattachment",
            name="events",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("registration_received", "registration received"),
                    ("registration_approved", "registration approved"),
                    ("registration_refused", "registration refused"),
                    ("registration_payment_request", "payment requested"),
                    ("registration_canceled", "registration canceled"),
                    ("discount_granted", "discount granted"),
                    ("payment_received", "payment received"),
                    ("payment_returned", "payment returned"),
                ],
                default=[],
                help_text="The attachment will be sent with notification on selected events.",
                max_length=166,
                verbose_name="send when",
            ),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="accounted_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="accounted by",
            ),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="amount",
            field=leprikon.models.fields.PriceField(
                decimal_places=0,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(Decimal("0.01"))],
                verbose_name="amount",
            ),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="transaction_type",
            field=models.CharField(
                choices=[
                    ("PAYMENT_CASH", "payment - cash"),
                    ("PAYMENT_BANK", "payment - bank"),
                    ("PAYMENT_ONLINE", "payment - online"),
                    ("RETURN_CASH", "returned payment - cash"),
                    ("RETURN_BANK", "returned payment - bank"),
                    ("TRANSFER", "transfer between registrations"),
                    ("DONATION_CASH", "donation - cash"),
                    ("DONATION_BANK", "donation - bank"),
                    ("DONATION_ONLINE", "donation - online"),
                    ("DONATION_TRANSFER", "transfer from registration to donation"),
                ],
                max_length=30,
                verbose_name="payment type",
            ),
        ),
    ]



from datetime import date
from itertools import chain
from traceback import print_exc

from django.conf import settings
from django.utils.translation import override
from django_cron import CronJobBase, Schedule
from sentry_sdk import capture_exception

from .models.courses import CourseRegistrationPeriod
from .models.events import EventRegistration
from .models.orderables import OrderableRegistration


class SentryCronJobBase(CronJobBase):
    def dojob(self):
        raise NotImplementedError(f"{self.__class__.__name__}.dojob must be implemented.")

    def do(self):
        try:
            with override(settings.LANGUAGE_CODE):
                return self.dojob()
        except Exception:
            capture_exception()
            raise


class SendPaymentRequest(SentryCronJobBase):
    schedule = Schedule(run_at_times=[settings.CRON_SEND_PAYMENT_REQUEST_TIME])
    code = "leprikon.cronjobs.SendPaymentRequest"

    def dojob(self):
        today = date.today()
        for registration in chain(
            set(
                registration_period.registration
                for registration_period in CourseRegistrationPeriod.objects.filter(
                    registration__approved__isnull=False,
                    registration__canceled__isnull=True,
                    payment_requested=False,
                    period__due_from__lte=today,
                ).select_related("registration")
            ),
            EventRegistration.objects.filter(
                approved__isnull=False,
                canceled__isnull=True,
                payment_requested__isnull=True,
                subject__event__due_from__lte=today,
            ),
            (
                registration
                for registration in OrderableRegistration.objects.filter(
                    approved__isnull=False,
                    canceled__isnull=True,
                    payment_requested__isnull=True,
                ).select_related("subject__orderable")
                if registration.get_due_from() <= today
            ),
        ):
            try:
                registration.request_payment(None)
            except Exception:
                print_exc()
                capture_exception()



from django import forms
from django.contrib import admin
from django.contrib.auth import get_user_model, login
from django.contrib.auth.decorators import permission_required
from django.contrib.messages import ERROR
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import path, reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from sentry_sdk import capture_exception
from user_unique_email.admin import UserAdmin as _UserAdmin

from ..forms.user import PasswordResetForm, UserAdminChangeForm, UserAdminCreateForm
from ..utils import attributes, merge_users
from .export import AdminExportMixin
from .messages import SendMessageAdminMixin

User = get_user_model()
admin.site.unregister(User)


class IsLeaderListFilter(admin.SimpleListFilter):
    parameter_name = "leprikon_leader__school_years"
    title = _("leader")

    def __init__(self, request, params, model, model_admin):
        self.school_year = request.school_year
        super().__init__(request, params, model, model_admin)

    def expected_parameters(self):
        return [self.parameter_name]

    def lookups(self, request, model_admin):
        return ((self.school_year.id, _("Yes")),)

    def queryset(self, request, queryset):
        value = self.value()
        if not value:
            return queryset
        return queryset.filter(leprikon_leader__school_years__id__exact=value)


@admin.register(User)
class UserAdmin(AdminExportMixin, SendMessageAdminMixin, _UserAdmin):
    actions = SendMessageAdminMixin.actions + AdminExportMixin.actions + ("merge", "reset_password")
    add_form = UserAdminCreateForm
    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("username", "first_name", "last_name", "email"),
            },
        ),
    )
    form = UserAdminChangeForm
    list_filter = _UserAdmin.list_filter + (IsLeaderListFilter,)

    @attributes(short_description=_("Merge selected user accounts"))
    def merge(self, request, queryset):
        class MergeForm(forms.Form):
            target = forms.ModelChoiceField(
                label=_("Target user"),
                help_text=_("All information will be merged into selected account."),
                queryset=queryset,
            )

        if request.POST.get("post", "no") == "yes":
            form = MergeForm(request.POST)
            if form.is_valid():
                target = form.cleaned_data["target"]
                for user in queryset.all():
                    if user != target:
                        try:
                            merge_users(user, target)
                            self.message_user(request, _("User {} was merged into user {}.").format(user, target))
                        except Exception:
                            self.message_user(
                                request, _("Can not merge user {} into user {}.").format(user, target), level=ERROR
                            )
                return
        else:
            form = MergeForm()
        return render(
            request,
            "leprikon/admin/merge.html",
            {
                "title": _("Select target user for merge"),
                "question": _(
                    "Are you sure you want to merge selected users into one? "
                    "All participants, parents, registrations and other related information "
                    "will be added to the target user account and the remaining users will be deleted."
                ),
                "queryset": queryset,
                "objects_title": _("Users"),
                "form_title": _("Select target account for merge"),
                "opts": self.model._meta,
                "form": form,
                "action_checkbox_name": admin.helpers.ACTION_CHECKBOX_NAME,
            },
        )

    @attributes(short_description=_("Send instructions to reset password to selected users"))
    def reset_password(self, request, queryset):
        form = PasswordResetForm()
        for user in queryset.all():
            try:
                form.cleaned_data = {"email": user.email}
                form.save(email_template_name="leprikon/password_reset_email.html")
                self.message_user(request, _("Instructions were send to user {}.").format(user))
            except Exception:
                capture_exception()
                self.message_user(request, _("Can not send instructions to user {}.").format(user), level=ERROR)

    def get_list_display(self, request):
        list_display = ["id"] + list(super().get_list_display(request))
        if request.user.is_superuser:
            list_display += ["login_as_link_superuser"]
        elif request.user.has_perm("auth.change_user"):
            list_display += ["login_as_link"]
        return list_display

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super().get_readonly_fields(request, obj)) + [
            "last_login",
            "date_joined",
        ]
        if not request.user.is_superuser:
            readonly_fields += ["is_staff", "is_superuser", "groups", "user_permissions"]
        return readonly_fields

    def get_urls(self):
        urls = super().get_urls()
        login_as_view = self.admin_site.admin_view(permission_required("auth.change_user")(self.login_as))
        return [
            path(
                "<int:user_id>/login-as/",
                login_as_view,
                name=f"{User._meta.app_label}_{User._meta.model_name}_login_as",
            )
        ] + urls

    def login_as(self, request, user_id):
        if request.user.is_superuser:
            user = get_object_or_404(get_user_model(), id=user_id)
        else:
            user = get_object_or_404(get_user_model(), id=user_id, is_staff=False, is_superuser=False)
        user.backend = "django.contrib.auth.backends.ModelBackend"
        login(request, user)
        return HttpResponseRedirect(reverse("leprikon:summary"))

    @attributes(short_description=_("login"))
    def login_as_link(self, obj):
        return mark_safe(
            '<a href="{url}">{text}</a>'.format(
                url=reverse(f"admin:{User._meta.app_label}_{User._meta.model_name}_login_as", args=[obj.id]),
                text=_("login"),
            )
            if not obj.is_staff and not obj.is_superuser
            else ""
        )

    @attributes(short_description=_("login"))
    def login_as_link_superuser(self, obj):
        return mark_safe(
            '<a href="{url}">{text}</a>'.format(
                url=reverse(f"admin:{User._meta.app_label}_{User._meta.model_name}_login_as", args=[obj.id]),
                text=_("login"),
            )
        )

    def get_message_recipients(self, request, queryset):
        return queryset.all()



from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from ..forms.subjects import RegistrationsFilterForm
from ..models.subjects import SubjectRegistration
from .generic import FilteredListView


class RegistrationsListView(FilteredListView):
    form_class = RegistrationsFilterForm
    message_empty = _("You don't have any registrations yet.")
    paginate_by = 10
    preview_template = "leprikon/registration_preview_object.html"
    title = _("My registrations")

    def get_queryset(self):
        qs = SubjectRegistration.objects.filter(
            user=self.request.user,
        ).order_by("-created")
        if not self.form.is_valid():
            return qs
        if self.form.cleaned_data.get("subject_types"):
            qs = qs.filter(subject__subject_type__in=self.form.cleaned_data["subject_types"])
        if self.form.cleaned_data.get("q"):
            for word in self.form.cleaned_data["q"].split():
                qs = qs.filter(Q(subject__name__icontains=word) | Q(subject__description__icontains=word))
        qs = qs.distinct()
        if self.form.cleaned_data.get("not_paid"):
            qs = [reg for reg in qs if reg.payment_status.amount_due]
        return qs



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-09-09 08:31
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0032_registration_participants_groups'),
    ]

    operations = [
        migrations.CreateModel(
            name='TargetGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='name')),
                ('order', models.IntegerField(blank=True, default=0, verbose_name='order')),
                ('require_school', models.BooleanField(default=True, verbose_name='require school')),
            ],
            options={
                'verbose_name': 'target group',
                'verbose_name_plural': 'target groups',
                'ordering': ('order',),
            },
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_city',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='city'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_email',
            field=leprikon.models.fields.EmailField(blank=True, default='', max_length=254, verbose_name='email address'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_first_name',
            field=models.CharField(blank=True, default='', max_length=30, verbose_name='first name'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_last_name',
            field=models.CharField(blank=True, default='', max_length=30, verbose_name='last name'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_phone',
            field=models.CharField(blank=True, default='', max_length=30, verbose_name='phone'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_postal_code',
            field=leprikon.models.fields.PostalCodeField(blank=True, default='', verbose_name='postal code'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_leader_street',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='street'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_school',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='leprikon.School', verbose_name='school'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_school_class',
            field=models.CharField(blank=True, default='', max_length=30, verbose_name='class'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='group_school_other',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='other school'),
        ),
        migrations.AlterField(
            model_name='subject',
            name='age_groups',
            field=models.ManyToManyField(blank=True, related_name='subjects', to='leprikon.AgeGroup', verbose_name='age groups'),
        ),
        migrations.AddField(
            model_name='subject',
            name='target_groups',
            field=models.ManyToManyField(blank=True, related_name='subjects', to='leprikon.TargetGroup', verbose_name='target groups'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='target_group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='leprikon.TargetGroup', verbose_name='target group'),
        ),
    ]



from typing import Optional

import pytest
from django.core.exceptions import ValidationError
from schwifty import IBAN

from .fields import parse_bank_account
from .utils import BankAccount


@pytest.mark.parametrize(
    "bank_account,expected_bank_account",
    [
        ("298182774/0300", BankAccount("CZ6303000000000298182774")),
        ("5083412002/5500", BankAccount("CZ2155000000005083412002")),
        ("670100-2200420359/6210", BankAccount("CZ7762106701002200420359")),
        ("5083412002 / 5500", BankAccount("CZ2155000000005083412002")),
        ("670100 - 2200420359 / 6210", BankAccount("CZ7762106701002200420359")),
        (IBAN("CZ6303000000000298182774"), BankAccount("CZ6303000000000298182774")),
        (BankAccount("CZ6303000000000298182774"), BankAccount("CZ6303000000000298182774")),
        (None, None),
    ],
)
def test_parse_bank_account(bank_account: Optional[str], expected_bank_account: BankAccount) -> None:
    """Given valid account number or None you get BankAccount or None."""

    assert parse_bank_account(bank_account) == expected_bank_account


@pytest.mark.parametrize(
    "invalid_bank_account",
    [
        "123456",
        "CZ6303000000000298182774",
        "CZ63 0300 0000 0002 9818 2774",
    ],
)
def test_parse_invalid_bank_account(invalid_bank_account: str) -> None:
    """Given invalid account number you get ValidationError."""
    with pytest.raises(ValidationError):
        parse_bank_account(invalid_bank_account)



from collections import defaultdict, namedtuple
from datetime import timedelta
from functools import lru_cache

from django.db.models import Sum
from django.template.response import TemplateResponse
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ...forms.reports.courses import CoursePaymentsForm, CoursePaymentsStatusForm, CourseStatsForm
from ...models.agegroup import AgeGroup
from ...models.citizenship import Citizenship
from ...models.courses import Course, CourseRegistration
from ...models.journals import JournalTime
from ...models.roles import Participant
from ...models.subjects import SubjectPayment, SubjectRegistrationParticipant, SubjectTime, SubjectType
from ...views.generic import FormView


class ReportCoursePaymentsView(FormView):
    form_class = CoursePaymentsForm
    template_name = "leprikon/reports/course_payments.html"
    title = _("Course payments")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["received_payments"] = SubjectPayment.objects.filter(
            target_registration__subject__subject_type__subject_type=SubjectType.COURSE,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["returned_payments"] = SubjectPayment.objects.filter(
            source_registration__subject__subject_type__subject_type=SubjectType.COURSE,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["received_payments_sum"] = context["received_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["returned_payments_sum"] = context["returned_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["sum"] = context["received_payments_sum"] - context["returned_payments_sum"]
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))


class ReportCoursePaymentsStatusView(FormView):
    form_class = CoursePaymentsStatusForm
    template_name = "leprikon/reports/course_payments_status.html"
    title = _("Course payments status")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    CoursePaymentsStatusSums = namedtuple("CoursePaymentsStatusSums", ("registrations", "status"))

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["reports"] = [
            self.Report(course, context["date"])
            for course in Course.objects.filter(school_year=self.request.school_year)
        ]
        context["sum"] = self.CoursePaymentsStatusSums(
            registrations=sum(len(r.registration_statuses) for r in context["reports"]),
            status=sum(r.status for r in context["reports"]),
        )
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))

    class Report:
        def __init__(self, course, d):
            self.course = course
            self.date = d

        RegPaymentStatus = namedtuple("RegPaymentStatus", ("registration", "status"))

        @cached_property
        def registration_statuses(self):
            return [
                registration_status
                for registration_status in (
                    self.RegPaymentStatus(
                        registration=registration,
                        status=registration.get_payment_status(self.date),
                    )
                    for registration in CourseRegistration.objects.filter(
                        subject=self.course,
                        approved__date__lte=self.date,
                    )
                )
                if registration_status.status.receivable
            ]

        @cached_property
        def status(self):
            return sum(rs.status for rs in self.registration_statuses)


class ReportCourseStatsView(FormView):
    form_class = CourseStatsForm
    template_name = "leprikon/reports/course_stats.html"
    title = _("Course statistics")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    ReportItem = namedtuple("ReportItem", ("age_group", "all", "boys", "girls", "citizenships"))

    _journal_deltas = {}
    _subject_deltas = {}

    def get_delta(self, cache, model, **kwargs):
        [id] = kwargs.values()
        try:
            return cache[id]
        except KeyError:
            cache[id] = sum(
                (t.delta for t in model.objects.filter(**kwargs)),
                start=timedelta(0),
            )
        return cache[id]

    @lru_cache
    def get_journal_delta(self, journal_id):
        return self.get_delta(self._journal_deltas, JournalTime, journal_id=journal_id)

    @lru_cache
    def get_subject_delta(self, subject_id):
        return self.get_delta(self._subject_deltas, SubjectTime, subject_id=subject_id)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["school_year"] = self.request.school_year
        return kwargs

    def form_valid(self, form):
        d = form.cleaned_data["date"]
        paid_only = form.cleaned_data["paid_only"]
        paid_later = form.cleaned_data["paid_later"]
        approved_later = form.cleaned_data["approved_later"]
        unique_participants = form.cleaned_data["unique_participants"]
        max_weekly_hours = form.cleaned_data["max_weekly_hours"]
        context = form.cleaned_data
        context["form"] = form

        if approved_later:
            # approved registrations created by the date
            participants = SubjectRegistrationParticipant.objects.filter(
                registration__created__date__lte=d,
                registration__approved__isnull=False,
            )
        else:
            # registrations approved by the date
            participants = SubjectRegistrationParticipant.objects.filter(
                registration__approved__date__lte=d,
            )
        participants = (
            participants.filter(
                registration__subject__in=form.cleaned_data["courses"],
            )
            .exclude(registration__canceled__date__lte=d)
            .select_related("registration", "age_group")
        )
        if paid_only:
            paid_date = None if paid_later else d
            participants = [
                participant
                for participant in participants
                if participant.registration.courseregistration.get_payment_status(paid_date).amount_due == 0
            ]
        else:
            participants = list(participants)

        context["courses_count"] = len(set(participant.registration.subject_id for participant in participants))

        weekly_delta_by_participant = defaultdict(timedelta)
        for participant in participants:
            weekly_delta_by_participant[participant.key] += sum(
                (self.get_journal_delta(journal.id) for journal in participant.journals.all()),
                start=timedelta(0),
            ) or self.get_subject_delta(participant.registration.subject_id)

        delta = sum(
            weekly_delta_by_participant.values(),
            start=timedelta(0),
        )
        context["participant_hours_count"] = delta.days * 24 + delta.seconds / 3600

        if unique_participants or max_weekly_hours:
            participants_by_key = {p.key: p for p in participants}
            if max_weekly_hours:
                max_weekly_delta = timedelta(hours=max_weekly_hours)
                participants = list(
                    p for p in participants_by_key.values() if weekly_delta_by_participant[p.key] <= max_weekly_delta
                )
            else:
                participants = list(participants_by_key.values())

        citizenships = list(Citizenship.objects.all())
        context["citizenships"] = citizenships

        context["participants_counts"] = self.ReportItem(
            age_group=None,
            all=len(participants),
            boys=len([p for p in participants if p.gender == Participant.MALE]),
            girls=len([p for p in participants if p.gender == Participant.FEMALE]),
            citizenships=[
                len([p for p in participants if p.citizenship_id == citizenship.id]) for citizenship in citizenships
            ],
        )
        context["participants_counts_by_age_groups"] = []
        for age_group in AgeGroup.objects.all():
            parts = [p for p in participants if p.age_group == age_group]
            context["participants_counts_by_age_groups"].append(
                self.ReportItem(
                    age_group=age_group,
                    all=len(parts),
                    boys=len([p for p in parts if p.gender == Participant.MALE]),
                    girls=len([p for p in parts if p.gender == Participant.FEMALE]),
                    citizenships=[
                        len([p for p in parts if p.citizenship_id == citizenship.id]) for citizenship in citizenships
                    ],
                )
            )

        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))



"""leprikon.site URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.sites.models import Site
from django.urls import include, path
from django.views.generic.base import RedirectView

try:
    urlpatterns = [
        path("favicon.ico", RedirectView.as_view(url="/static/favicon.ico")),
        path("admin/", admin.site.urls),
        path("pays/", include("django_pays.urls")),
        path("social/", include("social_django.urls")),
        path("verified-email-field/", include("verified_email_field.urls")),
        path("", include("leprikon.api.urls", namespace="api")),
        path("", include("cms.urls")),
        # this won't work for displaying pages,
        # but allows reverse resolving before leprikon apphook is attached to any page
        path("leprikon/", include("leprikon.urls")),
    ]
except Site.DoesNotExist:
    # this may happen during data migration
    print("Failed to load urls, because there is no Site.")
    urlpatterns = []

if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from typing import Any, List, Optional

from adminsortable2.admin import SortableAdminMixin, SortableInlineAdminMixin
from django import forms
from django.contrib import admin, messages
from django.contrib.admin.templatetags.admin_list import _boolean_icon
from django.contrib.admin.utils import unquote
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db.models import BooleanField, F, Func
from django.db.models.functions import Coalesce, Random
from django.http import HttpRequest, HttpResponseRedirect
from django.template.response import SimpleTemplateResponse
from django.templatetags.static import static
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..forms.subjects import (
    RegistrationAdminForm,
    RegistrationGroupAdminForm,
    RegistrationParticipantAdminForm,
    SubjectAdminForm,
    SubjectVariantForm,
)
from ..models.subjects import (
    DEFAULT_TEXTS,
    Subject,
    SubjectAttachment,
    SubjectGroup,
    SubjectReceivedPayment,
    SubjectRegistration,
    SubjectRegistrationBillingInfo,
    SubjectRegistrationGroup,
    SubjectRegistrationGroupMember,
    SubjectRegistrationParticipant,
    SubjectReturnedPayment,
    SubjectTime,
    SubjectType,
    SubjectTypeAttachment,
    SubjectVariant,
)
from ..models.utils import lazy_help_text_with_html_default
from ..utils import amount_color, attributes, currency
from .bulkupdate import BulkUpdateMixin
from .export import AdminExportMixin
from .filters import (
    ApprovedListFilter,
    CanceledListFilter,
    IsNullFieldListFilter,
    LeaderListFilter,
    SchoolYearListFilter,
    SubjectGroupListFilter,
    SubjectListFilter,
    SubjectTypeListFilter,
)
from .messages import SendMessageAdminMixin
from .sendmail import SendMailAdminMixin
from .transaction import TransactionAdminMixin, TransactionBaseAdmin, TransactionTypeListFilter
from .utils import datetime_with_by


class IsNull(Func):
    _output_field = BooleanField()
    arity = 1
    template = "%(expressions)s IS NULL"


class SubjectTypeAttachmentInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = SubjectTypeAttachment
    extra = 0


@admin.register(SubjectType)
class SubjectTypeAdmin(BulkUpdateMixin, SortableAdminMixin, admin.ModelAdmin):
    bulk_update_exclude = (
        "subject_type",
        "slug",
        "order",
        "name",
        "plural",
        "name_genitiv",
        "name_akuzativ",
    )
    list_display = ("plural",)
    filter_horizontal = ("questions", "registration_agreements")
    prepopulated_fields = {"slug": ("plural",)}
    inlines = (SubjectTypeAttachmentInlineAdmin,)

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)

        # limit choices of registration agreements
        registration_agreements_choices = form.base_fields["registration_agreements"].widget.widget.choices
        registration_agreements_choices.queryset = registration_agreements_choices.queryset.exclude(
            id__in=request.leprikon_site.registration_agreements.values("id")
        )
        form.base_fields["registration_agreements"].choices = registration_agreements_choices

        return form


@admin.register(SubjectGroup)
class SubjectGroupAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name", "color")
    list_editable = ("color",)
    filter_horizontal = ("subject_types",)


class SubjectAttachmentInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = SubjectAttachment
    extra = 0


class SubjectTimeInlineAdmin(admin.TabularInline):
    model = SubjectTime
    extra = 0


class SubjectBaseAdmin(AdminExportMixin, BulkUpdateMixin, SendMessageAdminMixin, admin.ModelAdmin):
    registration_model = None
    actions = (
        SendMessageAdminMixin.actions + BulkUpdateMixin.actions + AdminExportMixin.actions + ("set_registration_dates",)
    )
    list_editable = ("note",)
    list_filter = (
        ("school_year", SchoolYearListFilter),
        "department",
        ("subject_type", SubjectTypeListFilter),
        ("groups", SubjectGroupListFilter),
        ("leaders", LeaderListFilter),
        "place",
    )
    inlines = (
        SubjectTimeInlineAdmin,
        SubjectAttachmentInlineAdmin,
    )
    filter_horizontal = ("age_groups", "target_groups", "groups", "leaders", "questions", "registration_agreements")
    search_fields = ("name", "description")
    save_as = True

    class Media:
        js = ["leprikon/js/Popup.js"]

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and len(request.POST) == 4:
            return HttpResponseRedirect(
                "{}?subject_type={}&registration_type={}".format(
                    request.path,
                    request.POST.get("subject_type", ""),
                    request.POST.get("registration_type", ""),
                )
            )
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_exclude(self, request, obj=None):
        if request.registration_type == Subject.PARTICIPANTS:
            exclude = ["target_groups"]
        elif request.registration_type == Subject.GROUPS:
            exclude = ["age_groups"]
        else:
            exclude = []
        if obj and obj.registrations.exists():
            exclude.append("registration_type")
        return exclude

    def get_fieldsets(self, request, obj: Subject = None):
        available_fields = self.get_fields(request, obj)
        used_fields = []
        fieldsets = []
        for label, all_fields in [
            (
                None,
                [
                    "subject_type",
                    "registration_type",
                    "name",
                    "description",
                    # course specific
                    "school_year_division",
                    "allow_period_selection",
                    # event specific
                    "start_date",
                    "end_date",
                    "start_time",
                    "end_time",
                    "due_from",
                    "due_date",
                    # orderable specific
                    "duration",
                    "due_from_days",
                    "due_date_days",
                    "min_due_date_days",
                    "age_groups",
                    "target_groups",
                    "reg_from",
                    "reg_to",
                    "public",
                ],
            ),
            (
                _("advanced settings"),
                [
                    "code",
                    "department",
                    "groups",
                    "place",
                    "leaders",
                    "photo",
                    "page",
                    "min_registrations_count",
                    "max_registrations_count",
                    "note",
                    "questions",
                    "registration_agreements",
                    "reg_print_setup",
                    "decision_print_setup",
                    "pr_print_setup",
                    "bill_print_setup",
                    "organization",
                ],
            ),
            (
                _("texts"),
                [
                    "text_registration_received",
                    "text_registration_approved",
                    "text_registration_refused",
                    "text_registration_payment_request",
                    "text_registration_refund_offer",
                    "text_registration_canceled",
                    "text_discount_granted",
                    "text_payment_received",
                    "text_payment_returned",
                ],
            ),
        ]:
            fields = [f for f in all_fields if f in available_fields]
            if fields:
                fieldsets.append((label, {"fields": fields}))
                used_fields += fields
        unused_fields = [f for f in available_fields if f not in used_fields]
        if unused_fields:
            fieldsets.append((None, {"fields": unused_fields}))
        return fieldsets

    def get_form(self, request, obj, **kwargs):
        # set school year
        if obj:
            request.school_year = obj.school_year

        # get subject type
        try:
            # first try request.POST (user may want to change subject type)
            request.subject_type = SubjectType.objects.get(id=int(request.POST.get("subject_type")))
        except (SubjectType.DoesNotExist, TypeError, ValueError):
            if obj:
                # use subject type from object
                request.subject_type = obj.subject_type
            else:
                # try to get subject type from request.GET
                try:
                    request.subject_type = SubjectType.objects.get(
                        id=int(request.GET.get("subject_type")),
                    )
                except (SubjectType.DoesNotExist, TypeError, ValueError):
                    request.subject_type = None

        # get registration type
        request.registration_type = request.POST.get("registration_type")
        if request.registration_type not in Subject.REGISTRATION_TYPES:
            if obj:
                # use registration type from object
                request.registration_type = obj.registration_type
            else:
                # try to get registration type from request.GET
                request.registration_type = request.GET.get("registration_type")
                if request.registration_type not in Subject.REGISTRATION_TYPES:
                    request.registration_type = None

        if request.subject_type and request.registration_type:
            kwargs["form"] = type(
                SubjectAdminForm.__name__,
                (SubjectAdminForm,),
                {
                    "school_year": request.school_year,
                    "subject_type": request.subject_type,
                    "registration_type": request.registration_type,
                },
            )
        else:
            kwargs["fields"] = ["subject_type", "registration_type"]
            request.hide_inlines = True

        form = super().get_form(request, obj, **kwargs)

        if request.subject_type and request.registration_type:
            for field_name in [
                "text_registration_received",
                "text_registration_approved",
                "text_registration_refused",
                "text_registration_payment_request",
                "text_registration_refund_offer",
                "text_registration_canceled",
                "text_discount_granted",
                "text_payment_received",
                "text_payment_returned",
            ]:
                form.base_fields[field_name].help_text = lazy_help_text_with_html_default(
                    form.base_fields[field_name].help_text,
                    getattr(request.subject_type, field_name) or DEFAULT_TEXTS[field_name],
                )

        return form

    def get_inline_instances(self, request, obj=None):
        return [] if hasattr(request, "hide_inlines") else super().get_inline_instances(request, obj)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        formfield = super().formfield_for_foreignkey(db_field, request, **kwargs)
        if db_field.name == "subject_type":
            limit_choices_to = {"subject_type__exact": self.subject_type_type}
            formfield.limit_choices_to = limit_choices_to
        return formfield

    def save_form(self, request, form, change):
        obj = super().save_form(request, form, change)
        obj.school_year = request.school_year
        return obj

    @attributes(short_description=_("journals"))
    def journals_link(self, obj):
        return mark_safe(
            format_html(
                '<a href="{url}">{journals}</a>',
                url=reverse("admin:leprikon_journal_changelist")
                + f"?subject__subject_type__id__exact={obj.subject_type_id}&subject__id__exact={obj.id}",
                journals=_("journals"),
            )
            + format_html(
                '<a href="{url}" style="background-position: 0 0" title="{title}">' '<img src="{icon}" alt="+"/></a>',
                url=reverse("admin:leprikon_journal_add") + "?subject={}".format(obj.id),
                title=_("add journal"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )

    def get_message_recipients(self, request, queryset):
        return (
            get_user_model()
            .objects.filter(
                leprikon_registrations__canceled=None,
                leprikon_registrations__subject__in=queryset,
            )
            .distinct()
        )

    @attributes(short_description=_("registering"))
    def get_registering_link(self, obj):
        return mark_safe(
            format_html(
                '<a href="{url}">{subject_variants}</a>',
                url=reverse("admin:leprikon_subjectvariant_changelist") + f"?subject__exact={obj.id}",
                subject_variants=_("price and registering variants"),
            )
            + format_html(
                '<a href="{url}" style="background-position: 0 0" title="{title}">' '<img src="{icon}" alt="+"/></a>',
                url=reverse("admin:leprikon_subjectvariant_add") + "?subject={}".format(obj.id),
                title=_("add price and registering variant"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )

    @attributes(short_description=_("registrations"))
    def get_registrations_link(self, obj):
        icon = False
        approved_registrations_count = obj.approved_registrations.count()
        unapproved_registrations_count = obj.unapproved_registrations.count()

        if approved_registrations_count == 0:
            title = _("There are no approved registrations for this {}.").format(obj.subject_type.name_akuzativ)
        elif obj.min_registrations_count is not None and approved_registrations_count < obj.min_registrations_count:
            title = _("The number of approved registrations is lower than {}.").format(obj.min_registrations_count)
        elif obj.max_registrations_count is not None and approved_registrations_count > obj.max_registrations_count:
            title = _("The number of approved registrations is greater than {}.").format(obj.max_registrations_count)
        else:
            icon = True
            title = ""
        return mark_safe(
            format_html(
                '<a href="{url}" title="{title}">{icon} {approved}{unapproved}</a>',
                url=reverse(
                    "admin:{}_{}_changelist".format(
                        self.registration_model._meta.app_label,
                        self.registration_model._meta.model_name,
                    )
                )
                + "?subject__id__exact={}".format(obj.id),
                title=title,
                icon=_boolean_icon(icon),
                approved=approved_registrations_count,
                unapproved=" + {}".format(unapproved_registrations_count) if unapproved_registrations_count else "",
            )
            + format_html(
                '<a class="popup-link" href="{url}" style="background-position: 0 0" title="{title}">'
                '<img src="{icon}" alt="+"/></a>',
                url=reverse(
                    "admin:{}_{}_add".format(
                        self.registration_model._meta.app_label,
                        self.registration_model._meta.model_name,
                    )
                )
                + "?subject={}".format(obj.id),
                title=_("add registration"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )

    @attributes(short_description=_("approved registrations count"))
    def get_approved_registrations_count(self, obj):
        return obj.registrations.filter(canceled=None).exclude(approved=None).count()

    @attributes(short_description=_("unapproved registrations count"))
    def get_unapproved_registrations_count(self, obj):
        return obj.registrations.filter(canceled=None, approved=None).count()

    @attributes(short_description=_("photo"))
    def icon(self, obj):
        try:
            return mark_safe('<img src="{}" alt="{}"/>'.format(obj.photo.icons["48"], obj.photo.label))
        except (AttributeError, KeyError):
            return ""


class ChangeformRedirectMixin:
    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if object_id:
            obj = self.get_object(request, unquote(object_id))
            if obj:
                return HttpResponseRedirect(obj.get_edit_url())
        return super().changeform_view(request, object_id, form_url, extra_context)


@admin.register(Subject)
class SubjectAdmin(AdminExportMixin, SendMessageAdminMixin, ChangeformRedirectMixin, admin.ModelAdmin):
    """Hidden admin used for raw id fields"""

    actions = SendMessageAdminMixin.actions + AdminExportMixin.actions
    list_display = (
        "id",
        "code",
        "name",
        "subject_type",
        "get_groups_list",
        "get_leaders_list",
        "icon",
    )
    list_filter = (
        ("school_year", SchoolYearListFilter),
        "department",
        "subject_type__subject_type",
        ("subject_type", SubjectTypeListFilter),
        "registration_type",
        ("groups", SubjectGroupListFilter),
        ("leaders", LeaderListFilter),
    )
    search_fields = ("name", "description")

    def get_model_perms(self, request):
        return {}

    def has_add_permission(self, request, obj=None):
        return False

    @attributes(short_description=_("photo"))
    def icon(self, obj):
        try:
            return mark_safe('<img src="{}" alt="{}"/>'.format(obj.photo.icons["48"], obj.photo.label))
        except (AttributeError, KeyError):
            return ""


@admin.register(SubjectVariant)
class SubjectVariantAdmin(AdminExportMixin, BulkUpdateMixin, SendMessageAdminMixin, admin.ModelAdmin):
    actions = SendMessageAdminMixin.actions + BulkUpdateMixin.actions + AdminExportMixin.actions
    bulk_update_exclude = ("subject", "order")
    list_display = (
        "subject",
        "name",
        "reg_from",
        "reg_to",
        "registration_price",
        "participant_price",
        "school_year_division",
        "get_registrations_link",
        "order",
    )
    list_editable = ("order",)
    list_filter = (
        ("subject__school_year", SchoolYearListFilter),
        "subject__department",
        "subject__subject_type__subject_type",
        ("subject__subject_type", SubjectTypeListFilter),
        ("subject", SubjectListFilter),
    )
    raw_id_fields = ("subject",)
    search_fields = ("name", "description")
    filter_horizontal = ("age_groups", "target_groups")

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and "subject" in request.POST:
            return HttpResponseRedirect(f"{request.path}?subject={request.POST['subject']}")
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_form(self, request: HttpRequest, obj: SubjectVariant | None, **kwargs):
        if obj:
            request.school_year = obj.subject.school_year
            request.subject = obj.subject
        else:
            # try to get subject from request.GET
            try:
                request.subject = Subject.objects.get(
                    school_year=request.school_year,
                    id=int(request.GET.get("subject")),
                )
            except (Subject.DoesNotExist, TypeError, ValueError):
                request.subject = None

        if request.subject:
            kwargs["form"] = type(SubjectVariantForm.__name__, (SubjectVariantForm,), {"subject": request.subject})
        else:
            kwargs["fields"] = ["subject"]

        return super().get_form(request, obj, **kwargs)

    def get_bulk_update_form(self, request: HttpRequest, fields: List[str]):
        form = super().get_bulk_update_form(request, fields)
        if "school_year_division" in form.base_fields:
            form = type(
                form.__name__,
                (form,),
                {"school_year_division": forms.ModelChoiceField(request.school_year.divisions.all())},
            )
        return form

    def get_exclude(self, request: HttpRequest, obj: SubjectVariant | None) -> Any:
        if not request.subject:
            return []
        exclude = ["subject", "order"]
        if request.subject:
            if request.subject.registration_type_participants:
                exclude.append("target_groups")
            if request.subject.registration_type_groups:
                exclude.append("age_groups")
            if request.subject.subject_type.subject_type != SubjectType.COURSE:
                exclude.append("school_year_division")
        return exclude

    def get_message_recipients(self, request, queryset):
        return (
            get_user_model()
            .objects.filter(
                leprikon_registrations__canceled=None,
                leprikon_registrations__subject_variant__in=queryset,
            )
            .distinct()
        )

    @attributes(short_description=_("registrations"))
    def get_registrations_link(self, obj):
        approved_registrations_count = obj.approved_registrations.count()
        unapproved_registrations_count = obj.unapproved_registrations.count()
        registration_model = {
            SubjectType.COURSE: "courseregistration",
            SubjectType.EVENT: "eventregistration",
            SubjectType.ORDERABLE: "orderableregistration",
        }[obj.subject.subject_type.subject_type]

        return mark_safe(
            format_html(
                '<a href="{url}">{approved}{unapproved}</a>',
                url=reverse(f"admin:leprikon_{registration_model}_changelist")
                + "?subject__id__exact={}&subject_variant__id__exact={}".format(obj.subject_id, obj.id),
                approved=approved_registrations_count,
                unapproved=" + {}".format(unapproved_registrations_count) if unapproved_registrations_count else "",
            )
            + format_html(
                '<a class="popup-link" href="{url}" style="background-position: 0 0" title="{title}">'
                '<img src="{icon}" alt="+"/></a>',
                url=reverse(f"admin:leprikon_{registration_model}_add")
                + "?subject={}&subject_variant={}".format(obj.subject_id, obj.id),
                title=_("add registration"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )


class SubjectRegistrationParticipantInlineAdmin(admin.StackedInline):
    model = SubjectRegistrationParticipant
    extra = 0

    def get_min_num(self, request, obj=None, **kwargs):
        return request.subject_variant.min_participants_count

    def get_max_num(self, request, obj=None, **kwargs):
        return request.subject_variant.max_participants_count

    def get_formset(self, request, obj=None, **kwargs):
        questions = obj.all_questions if obj else request.subject.all_questions
        fields = dict(("q_" + q.slug, q.get_field()) for q in questions)
        fields["subject"] = request.subject
        fields["obj"] = obj
        kwargs["form"] = type(RegistrationParticipantAdminForm.__name__, (RegistrationParticipantAdminForm,), fields)
        return super().get_formset(request, obj, **kwargs)


class SubjectRegistrationGroupInlineAdmin(admin.StackedInline):
    model = SubjectRegistrationGroup
    extra = 1
    min_num = 1
    max_num = 1

    def get_formset(self, request, obj=None, **kwargs):
        questions = obj.all_questions if obj else request.subject.all_questions
        fields = dict(("q_" + q.slug, q.get_field()) for q in questions)
        fields["subject"] = request.subject
        fields["obj"] = obj
        kwargs["form"] = type(RegistrationGroupAdminForm.__name__, (RegistrationGroupAdminForm,), fields)
        return super().get_formset(request, obj, **kwargs)


class SubjectRegistrationGroupMemberInlineAdmin(admin.StackedInline):
    model = SubjectRegistrationGroupMember
    extra = 0

    def get_min_num(self, request, obj=None, **kwargs):
        return request.subject_variant.min_participants_count

    def get_max_num(self, request, obj=None, **kwargs):
        return request.subject_variant.max_participants_count


class RegistrationBillingInfoInlineAdmin(admin.TabularInline):
    model = SubjectRegistrationBillingInfo
    max_num = 1
    extra = 0


class SubjectRegistrationBaseAdmin(AdminExportMixin, SendMailAdminMixin, SendMessageAdminMixin, admin.ModelAdmin):
    actions = (
        SendMessageAdminMixin.actions
        + SendMailAdminMixin.actions
        + AdminExportMixin.actions
        + (
            "approve",
            "refuse",
            "request_payment",
            "offer_refund",
            "generate_refund_request",
            "export_invoices_xml",
            "cancel",
            "cancel_cancelation_request",
        )
    )
    form = RegistrationAdminForm
    inlines = (RegistrationBillingInfoInlineAdmin,)
    list_display = (
        "variable_symbol",
        "download_tag",
        "decision",
        "subject_name",
        "participants_list_html",
        "price",
        "discounts",
        "total_price",
        "received_payments",
        "returned_payments",
        "created_with_by",
        "approved_with_by",
        "payment_requested_with_by",
        "refund_offered_with_by",
        "cancelation_requested_with_by",
        "canceled_with_by",
        "note",
        "random_number",
    )
    list_editable = ("note",)
    list_export = (
        "id",
        "variable_symbol",
        "slug",
        "user",
        "subject",
        "subject_variant",
        "price",
        "discounts_export",
        "total_price_export",
        "received_payments_export",
        "note",
        "created",
        "created_by",
        "payment_requested",
        "payment_requested_by",
        "refund_offered",
        "refund_offered_by",
        "approved",
        "approved_by",
        "cancelation_requested",
        "cancelation_requested_by",
        "canceled",
        "canceled_by",
        "agreement_options_list",
        "group_members_list",
        "participants__gender",
        "participants__first_name",
        "participants__last_name",
        "participants__birth_num",
        "participants__birth_date",
        "participants__gender",
        "participants__age_group",
        "participants__street",
        "participants__city",
        "participants__postal_code",
        "participants__citizenship",
        "participants__phone",
        "participants__email",
        "participants__school",
        "participants__school_other",
        "participants__school_class",
        "participants__health",
        "participants__answers",
        "participants__has_parent1",
        "participants__parent1_first_name",
        "participants__parent1_last_name",
        "participants__parent1_street",
        "participants__parent1_city",
        "participants__parent1_postal_code",
        "participants__parent1_phone",
        "participants__parent1_email",
        "participants__has_parent2",
        "participants__parent2_first_name",
        "participants__parent2_last_name",
        "participants__parent2_street",
        "participants__parent2_city",
        "participants__parent2_postal_code",
        "participants__parent2_phone",
        "participants__parent2_email",
        "group__name",
        "group__first_name",
        "group__last_name",
        "group__street",
        "group__city",
        "group__postal_code",
        "group__phone",
        "group__email",
        "group__school__name",
        "group__school_other",
        "group__school_class",
        "billing_info__name",
        "billing_info__street",
        "billing_info__city",
        "billing_info__postal_code",
        "billing_info__company_num",
        "billing_info__vat_number",
        "billing_info__contact_person",
        "billing_info__phone",
        "billing_info__email",
        "billing_info__employee",
    )
    list_filter = (
        ("subject__school_year", SchoolYearListFilter),
        "subject__department",
        ("subject__subject_type", SubjectTypeListFilter),
        "subject__registration_type",
        "subject__organization",
        ApprovedListFilter,
        CanceledListFilter,
        "registration_link",
        ("billing_info", IsNullFieldListFilter),
        "subject__groups",
        ("subject", SubjectListFilter),
        ("subject__leaders", LeaderListFilter),
        "subject__place",
    )
    search_fields = (
        "variable_symbol",
        "participants__birth_num",
        "participants__first_name",
        "participants__last_name",
        "participants__parent1_first_name",
        "participants__parent1_last_name",
        "participants__parent2_first_name",
        "participants__parent2_last_name",
        "group_members__first_name",
        "group_members__last_name",
    )
    ordering = ("-created",)
    raw_id_fields = (
        "subject",
        "user",
    )

    class Media:
        css = {"all": ["leprikon/css/registrations.changelist.css"]}
        js = ["leprikon/js/Popup.js"]

    def has_delete_permission(self, request, obj=None):
        if obj and obj.approved is None and obj.received_payments.count() == 0 and obj.returned_payments.count() == 0:
            return super().has_delete_permission(request, obj)
        else:
            return False

    def lookup_allowed(self, lookup: str, value: str) -> bool:
        return lookup in (
            "participants__birth_date",
            "participants__birth_num",
            "participants__first_name",
            "participants__last_name",
        ) or super().lookup_allowed(lookup, value)

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def get_changelist(self, request, **kwargs):
        class ChangeList(super().get_changelist(request, **kwargs)):
            def get_ordering(self, request, queryset):
                # Show registrations with cancelation request on the top
                # if not showing canceled ones.
                ordering = super().get_ordering(request, queryset)
                return (
                    ordering
                    if request.GET.get("canceled") == "yes"
                    else [
                        IsNull("cancelation_requested"),
                        *ordering,
                    ]
                )

        return ChangeList

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related(
                "discounts",
                "received_payments",
                "returned_payments",
            )
            .select_related(
                "subject",
                "user",
            )
            .annotate(
                random_number=Random(),
                subject__organization__id=Coalesce(
                    F("subject__organization_id"),
                    F("subject__subject_type__organization_id"),
                ),
            )
        )

    @attributes(short_description=_("Approve selected registrations"))
    def approve(self, request, queryset):
        for registration in queryset.all():
            try:
                registration.approve(request.user)
            except ValidationError as e:
                self.message_user(request, e.message, messages.ERROR)
            else:
                self.message_user(
                    request,
                    _("The registration {r} has been approved and the user has been notified.").format(r=registration),
                )

    @attributes(short_description=_("Refuse selected registrations"))
    def refuse(self, request, queryset):
        for registration in queryset.all():
            try:
                registration.refuse(request.user)
            except ValidationError as e:
                self.message_user(request, e.message, messages.ERROR)
            else:
                self.message_user(
                    request,
                    _("The registration {r} has been refused and the user has been notified.").format(r=registration),
                )

    @attributes(short_description=_("Request payment for selected registrations"))
    def request_payment(self, request, queryset):
        for registration in queryset.all():
            registration.request_payment(request.user)
        self.message_user(request, _("Payment was requested for selected registrations."))

    @attributes(short_description=_("Offer refund for selected registrations"))
    def offer_refund(self, request, queryset):
        for registration in queryset.all():
            registration.offer_refund(request.user)
        self.message_user(request, _("Refund was offered for selected registrations."))

    @attributes(short_description=_("Generate refund requests for selected registrations"))
    def generate_refund_request(self, request, queryset):
        for registration in queryset.filter(refund_request__bank_account=None):
            registration.generate_refund_request(request.user)
        self.message_user(request, _("Refund requests were generated for selected registrations."))

    @attributes(short_description=_("Export selected registrations as invoices in XML"))
    def export_invoices_xml(self, request, queryset):
        response = SimpleTemplateResponse("leprikon/invoices.xml", {"registrations": queryset}, content_type="text/xml")
        response["Content-Disposition"] = 'attachment; filename="invoices.xml"'
        return response

    @attributes(short_description=_("Cancel selected registrations"))
    def cancel(self, request, queryset):
        for registration in queryset.all():
            try:
                registration.cancel(request.user)
            except ValidationError as e:
                self.message_user(request, e.message, messages.ERROR)
            else:
                self.message_user(
                    request,
                    _("The registration {r} has been canceled and the user has been notified.").format(r=registration),
                )

    @attributes(short_description=_("Cancel cancelation request for selected registrations"))
    def cancel_cancelation_request(self, request, queryset):
        queryset.update(cancelation_requested=None, cancelation_requested_by=None)
        self.message_user(request, _("The cancelation requests have been removed from selected registrations."))

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        formfield = super().formfield_for_foreignkey(db_field, request, **kwargs)
        if db_field.name == "subject":
            limit_choices_to = {"subject_type__subject_type__exact": self.model.subject_type}
            formfield.limit_choices_to = limit_choices_to
            formfield.widget.rel.limit_choices_to = limit_choices_to
        return formfield

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and "user" not in request.POST:
            return HttpResponseRedirect(
                "{}?subject={}&subject_variant={}".format(
                    request.path, request.POST.get("subject", ""), request.POST.get("subject_variant", "")
                )
            )
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_exclude(self, request, obj=None):
        return [] if not request.subject or request.subject.variants.exists() else ["subject_variant"]

    def get_form(self, request, obj: Optional[SubjectRegistration], **kwargs):
        try:
            # first try request.POST (user may want to change the subject)
            request.subject = Subject.objects.get(id=int(request.POST.get("subject")))
        except (Subject.DoesNotExist, TypeError, ValueError):
            if obj:
                # use subject from object
                request.subject = obj.subject
            else:
                # try to get subject from request.GET
                try:
                    request.subject = Subject.objects.get(id=int(request.GET.get("subject")))
                except (Subject.DoesNotExist, TypeError, ValueError):
                    request.subject = None

        if request.subject:
            try:
                # first try request.POST (user may want to change the subject variant)
                request.subject_variant = SubjectVariant.objects.get(
                    id=int(request.POST.get("subject_variant")),
                    subject=request.subject,
                )
            except (SubjectVariant.DoesNotExist, TypeError, ValueError):
                if obj and obj.subject_variant.subject == request.subject:
                    # use subject from object
                    request.subject_variant = obj.subject_variant
                else:
                    # try to get subject variant from request.GET
                    try:
                        request.subject_variant = SubjectVariant.objects.get(
                            id=int(request.GET.get("subject_variant")),
                            subject=request.subject,
                        )
                    except (SubjectVariant.DoesNotExist, TypeError, ValueError):
                        request.subject_variant = None

        if request.subject:
            if request.subject_variant:
                kwargs["form"] = type(
                    self.form.__name__,
                    (self.form,),
                    {"subject": request.subject, "subject_variant": request.subject_variant},
                )
            else:
                kwargs["form"] = type(
                    RegistrationAdminForm.__name__, (RegistrationAdminForm,), {"subject": request.subject}
                )
                kwargs["fields"] = ["subject", "subject_variant"]
        else:
            kwargs["form"] = forms.ModelForm
            kwargs["fields"] = ["subject"]
        return super().get_form(request, obj, **kwargs)

    def get_inline_instances(self, request, obj=None):
        if request.subject and request.subject_variant:
            if request.subject.registration_type_participants:
                inlines = (SubjectRegistrationParticipantInlineAdmin,)
            elif request.subject.registration_type_groups:
                inlines = (SubjectRegistrationGroupInlineAdmin, SubjectRegistrationGroupMemberInlineAdmin)
            return [inline(self.model, self.admin_site) for inline in inlines] + super().get_inline_instances(
                request, obj
            )
        else:
            return []

    def save_model(self, request, obj, form, change):
        if not change:
            if obj.subject.registration_type_participants:
                participants_count = int(form.data["participants-TOTAL_FORMS"])
            elif obj.subject.registration_type_groups:
                participants_count = int(form.data["group_members-TOTAL_FORMS"])
            obj.price = obj.subject_variant.get_price(participants_count)
            obj.created_by = request.user
        super().save_model(request, obj, form, change)

    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)
        if not change:
            form.instance.questions.set(form.instance.subject.all_questions)
            form.instance.agreements.set(form.instance.subject.all_registration_agreements)
            form.instance.generate_variable_symbol_and_slug()
            form.instance.send_mail()

    def get_css(self, obj):
        classes = []
        if obj.cancelation_requested and not obj.canceled:
            classes.append("reg-cancel-request")
        if obj.approved:
            classes.append("reg-approved")
        else:
            classes.append("reg-new")
            if obj.subject.full:
                classes.append("reg-full")
        if obj.canceled:
            classes.append("reg-canceled")
        else:
            classes.append("reg-active")
        return " ".join(classes)

    legend = (
        ("reg-cancel-request", _("cancelation requested")),
        ("reg-new", _("new registration")),
        ("reg-new reg-full", _("new registration, capacity full")),
        ("reg-approved", _("approved registration")),
    )

    @attributes(short_description=_("subject"))
    def subject_name(self, obj):
        return obj.subject.name

    @attributes(short_description=_("decision"))
    def decision(self, obj):
        if obj.approved or obj.canceled:
            return mark_safe(
                '<a href="{}">{}</a>'.format(
                    reverse(
                        "admin:{}_{}_event_pdf".format(self.model._meta.app_label, self.model._meta.model_name),
                        args=(obj.id, "decision"),
                    ),
                    _("admission decision") if obj.approved else _("refusal decision"),
                )
            )

    created_with_by = datetime_with_by("created", _("time of registration"))

    approved_with_by = datetime_with_by("approved", _("time of approval"))

    payment_requested_with_by = datetime_with_by("payment_requested", _("payment request time"))

    refund_offered_with_by = datetime_with_by("refund_offered", _("refund offer time"))

    cancelation_requested_with_by = datetime_with_by("cancelation_requested", _("time of cancellation request"))

    canceled_with_by = datetime_with_by("canceled", _("time of cancellation"))

    def get_message_recipients(self, request, queryset):
        return get_user_model().objects.filter(leprikon_registrations__in=queryset).distinct()

    @attributes(admin_order_field="random_number", short_description=_("random number"))
    def random_number(self, obj):
        return int(obj.random_number * 1000000000000)

    @attributes(short_description=_("received payments"))
    def received_payments(self, obj: SubjectRegistration):
        return mark_safe(
            format_html(
                '<a style="color: {color}" href="{href_list}" title="{title}"><b>{amount}</b></a>'
                ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
                '<img src="{icon_add}" alt="+"/></a>',
                amount=currency(obj.payment_status.received),
                color=obj.payment_status.color,
                href_add=reverse("admin:leprikon_subjectreceivedpayment_add") + f"?target_registration={obj.id}",
                href_list=reverse("admin:leprikon_subjectreceivedpayment_changelist")
                + f"?target_registration={obj.id}",
                icon_add=static("admin/img/icon-addlink.svg"),
                title=obj.payment_status.title,
                title_add=_("add received payment"),
            )
        )

    @attributes(short_description=_("returned payments"))
    def returned_payments(self, obj: SubjectRegistration):
        return mark_safe(
            format_html(
                '<a href="{href_list}"><b>{amount}</b></a>'
                ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
                '<img src="{icon_add}" alt="+"/></a>',
                amount=currency(obj.payment_status.returned),
                href_add=reverse("admin:leprikon_subjectreturnedpayment_add") + f"?source_registration={obj.id}",
                href_list=reverse("admin:leprikon_subjectreturnedpayment_changelist")
                + f"?source_registration={obj.id}",
                icon_add=static("admin/img/icon-addlink.svg"),
                title_add=_("add returned payment"),
            )
        )

    @attributes(short_description=_("total price"))
    def total_price(self, obj: SubjectRegistration):
        return currency(obj.payment_status.receivable)


@admin.register(SubjectRegistration)
class SubjectRegistrationAdmin(AdminExportMixin, SendMessageAdminMixin, ChangeformRedirectMixin, admin.ModelAdmin):
    """Hidden admin used for raw id fields"""

    actions = SendMessageAdminMixin.actions + AdminExportMixin.actions
    list_display = (
        "id",
        "variable_symbol",
        "subject",
        "participants_list_html",
        "group",
        "created",
        "canceled",
    )
    list_filter = (
        ("subject__school_year", SchoolYearListFilter),
        "subject__department",
        ("subject__subject_type", SubjectTypeListFilter),
        ApprovedListFilter,
        CanceledListFilter,
        ("subject", SubjectListFilter),
        ("subject__leaders", LeaderListFilter),
    )
    ordering = ("-created",)
    search_fields = (
        "variable_symbol",
        "participants__birth_num",
        "participants__first_name",
        "participants__last_name",
        "participants__parent1_first_name",
        "participants__parent1_last_name",
        "participants__parent2_first_name",
        "participants__parent2_last_name",
        "group_members__first_name",
        "group_members__last_name",
    )

    def lookup_allowed(self, lookup: str, value: str) -> bool:
        return lookup in (
            "participants__birth_date",
            "participants__birth_num",
            "participants__first_name",
            "participants__last_name",
        ) or super().lookup_allowed(lookup, value)

    def get_model_perms(self, request):
        return {}

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_actions(self, request):
        return {}


class SubjectDiscountBaseAdmin(TransactionAdminMixin, AdminExportMixin, admin.ModelAdmin):
    list_filter = (
        ("registration__subject__school_year", SchoolYearListFilter),
        "registration__subject__department",
        ("registration__subject__subject_type", SubjectTypeListFilter),
        ("registration__subject", SubjectListFilter),
        ("registration__subject__leaders", LeaderListFilter),
    )
    search_fields = (
        "registration__variable_symbol",
        "registration__subject__name",
        "registration__participants__first_name",
        "registration__participants__last_name",
        "registration__participants__birth_num",
    )
    raw_id_fields = ("registration",)
    closed_fields = ("accounted", "registration", "amount")

    @attributes(short_description=_("subject"))
    def subject(self, obj):
        return obj.registration.subject

    list_export = (
        "id",
        "accounted",
        "accounted_by",
        "transaction_type_label",
        "amount",
        "organization",
        "donor__username",
        "donor",
        "last_updated",
        "last_updated_by",
        "mail_sent",
        "note",
    )


class SubjectPaymentAdminMixin:
    list_display = (
        "id",
        "accounted_with_by",
        "download_tag",
        "transaction_type",
        "amount_html",
        "registration",
        "last_updated_with_by",
        "mail_sent",
        "note",
    )
    list_export = (
        "id",
        "accounted",
        "accounted_by",
        "transaction_type_label",
        "amount",
        "registration",
        "variable_symbol",
        "last_updated",
        "last_updated_by",
        "note",
    )

    @attributes(short_description=_("variable symbol"))
    def variable_symbol(self, obj):
        return obj.registration.variable_symbol


@admin.register(SubjectReceivedPayment)
class SubjectReceivedPaymentAdmin(SubjectPaymentAdminMixin, TransactionBaseAdmin):
    exclude = ("donor", "organization")
    list_filter = (
        ("target_registration__subject__school_year", SchoolYearListFilter),
        ("transaction_type", TransactionTypeListFilter),
        "target_registration__subject__department",
        ("target_registration__subject__subject_type", SubjectTypeListFilter),
        ("target_registration__subject", SubjectListFilter),
        ("target_registration__subject__leaders", LeaderListFilter),
    )
    search_fields = (
        "target_registration__subject__name",
        "target_registration__participants__first_name",
        "target_registration__participants__last_name",
        "target_registration__participants__birth_num",
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.base_fields["target_registration"].required = True
        return form


@admin.register(SubjectReturnedPayment)
class SubjectReturnedPaymentAdmin(SubjectPaymentAdminMixin, TransactionBaseAdmin):
    exclude = ("pays_payment",)
    list_filter = (
        ("source_registration__subject__school_year", SchoolYearListFilter),
        ("transaction_type", TransactionTypeListFilter),
        "source_registration__subject__department",
        ("source_registration__subject__subject_type", SubjectTypeListFilter),
        ("source_registration__subject", SubjectListFilter),
        ("source_registration__subject__leaders", LeaderListFilter),
    )
    search_fields = (
        "source_registration__subject__name",
        "source_registration__participants__first_name",
        "source_registration__participants__last_name",
        "source_registration__participants__birth_num",
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.base_fields["source_registration"].required = True
        return form

    @attributes(admin_order_field="amount", short_description=_("amount"))
    def amount_html(self, obj):
        return mark_safe(
            format_html(
                '<b style="color: {color}">{amount}</b>',
                color=amount_color(-obj.amount),
                amount=currency(obj.amount),
            )
        )



from django.utils.translation import gettext_lazy as _

"""
    This module contains strings from other packages, which are not correctly
    translated to Czech language.
"""

# cms
_("Publish page changes")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-07-12 08:05
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0022_auto_20180620_1551'),
        ('leprikon', '0054_registration_dates_by_cleanup'),
    ]

    operations = [
        migrations.CreateModel(
            name='SchoolYearBlockPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='leprikon_schoolyearblockplugin', serialize=False, to='cms.CMSPlugin')),
                ('school_years', models.ManyToManyField(help_text='Which school years should the block be visible for?', related_name='_schoolyearblockplugin_school_years_+', to='leprikon.SchoolYear', verbose_name='school years')),
            ],
            bases=('cms.cmsplugin',),
        ),
    ]



from django import forms
from django.forms import MultiWidget
from django.forms.widgets import CheckboxSelectMultiple, RadioSelect
from django.templatetags.static import static


class RadioSelectBootstrap(RadioSelect):
    template_name = "leprikon/widgets/multiple_input.html"
    option_template_name = "leprikon/widgets/input_option.html"


class CheckboxSelectMultipleBootstrap(CheckboxSelectMultiple):
    template_name = "leprikon/widgets/multiple_input.html"
    option_template_name = "leprikon/widgets/input_option.html"


class RequiredOptionalWidget(MultiWidget):
    template_name = "required_optional_field.html"

    def __init__(self, widgets, attrs, empty_value="-") -> None:
        self.empty_value = empty_value
        super().__init__(widgets, attrs)

    @property
    def media(self):
        return forms.Media(js=[static("leprikon/js/required_optional_field.js")])

    def decompress(self, value):
        if value is None:
            return (None, None)
        if value == self.empty_value:
            return ("empty", None)
        return ("value", value)



# Generated by Django 3.2.16 on 2023-01-08 18:45

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("django_pays", "0001_initial"),
        ("leprikon", "0072_prices"),
    ]

    operations = [
        migrations.AlterField(
            model_name="courselistplugin",
            name="course_types",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by course types.",
                limit_choices_to={"subject_type": "course"},
                related_name="_leprikon_courselistplugin_course_types_+",
                to="leprikon.SubjectType",
                verbose_name="course types",
            ),
        ),
        migrations.AlterField(
            model_name="eventlistplugin",
            name="event_types",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by event types.",
                limit_choices_to={"subject_type": "event"},
                related_name="_leprikon_eventlistplugin_event_types_+",
                to="leprikon.SubjectType",
                verbose_name="event types",
            ),
        ),
        migrations.AlterField(
            model_name="filteredcourselistplugin",
            name="course_types",
            field=models.ManyToManyField(
                limit_choices_to={"subject_type": "course"},
                related_name="_leprikon_filteredcourselistplugin_course_types_+",
                to="leprikon.SubjectType",
                verbose_name="course types",
            ),
        ),
        migrations.AlterField(
            model_name="filteredeventlistplugin",
            name="event_types",
            field=models.ManyToManyField(
                limit_choices_to={"subject_type": "event"},
                related_name="_leprikon_filteredeventlistplugin_event_types_+",
                to="leprikon.SubjectType",
                verbose_name="event types",
            ),
        ),
        migrations.AlterField(
            model_name="filteredorderablelistplugin",
            name="event_types",
            field=models.ManyToManyField(
                limit_choices_to={"subject_type": "orderable"},
                related_name="_leprikon_filteredorderablelistplugin_event_types_+",
                to="leprikon.SubjectType",
                verbose_name="event types",
            ),
        ),
        migrations.AlterField(
            model_name="message",
            name="sender",
            field=models.ForeignKey(
                limit_choices_to={"is_staff": True},
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="sender",
            ),
        ),
        migrations.AlterField(
            model_name="orderablelistplugin",
            name="event_types",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by event types.",
                limit_choices_to={"subject_type": "orderable"},
                related_name="_leprikon_orderablelistplugin_event_types_+",
                to="leprikon.SubjectType",
                verbose_name="event types",
            ),
        ),
    ]



from datetime import date

from django.contrib.auth import authenticate, login, logout
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import AuthenticationFailed, NotFound
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response

from ..models.journals import Journal
from ..models.schoolyear import SchoolYear
from .serializers import (
    CredentialsSerializer,
    SchoolYearSerializer,
    SetSchoolYearSerializer,
    SubjectRegistrationParticipantSerializer,
    UserSerializer,
)


class JournalViewSet(viewsets.GenericViewSet):
    def get_queryset(self):
        if self.request.user.is_staff:
            return Journal.objects.all()
        return Journal.objects.filter(leaders=self.request.leader)

    @extend_schema(
        responses={200: SubjectRegistrationParticipantSerializer(many=True)},
        methods=["get"],
        parameters=[OpenApiParameter(name="date", type=OpenApiTypes.DATE)],
    )
    @action(detail=True, methods=["get"], permission_classes=[IsAuthenticated])
    def participants(self, request, pk: str):
        journal: Journal = self.get_object()
        try:
            d = date.fromisoformat(request.GET["date"])
        except (KeyError, ValueError):
            d = date.today()

        participants = journal.get_valid_participants(d)

        return Response(SubjectRegistrationParticipantSerializer(participants, many=True).data)


class SchoolYearViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = SchoolYearSerializer

    def get_queryset(self):
        if self.request.user and self.request.user.is_staff:
            return SchoolYear.objects.all()
        return SchoolYear.objects.filter(active=True)

    @extend_schema(request=None, responses={200: SchoolYearSerializer}, methods=["get"])
    @extend_schema(request=SetSchoolYearSerializer, responses={204: None}, methods=["post"])
    @action(detail=False, methods=["get", "post"])
    def current(self, request):
        if request.method == "GET":
            return Response(SchoolYearSerializer(request.school_year).data)

        if request.method == "POST":
            serializer = SetSchoolYearSerializer(data=request.data)
            serializer.is_valid(raise_exception=True)
            try:
                request._request.school_year = self.get_queryset().get(pk=serializer.validated_data["id"])
                return Response(status=status.HTTP_204_NO_CONTENT)
            except SchoolYear.DoesNotExist:
                raise NotFound


class UserViewSet(viewsets.ViewSet):
    @extend_schema(operation_id="user_login", request=CredentialsSerializer, responses={200: UserSerializer})
    @action(detail=False, methods=["post"])
    def login(self, request):
        serializer = CredentialsSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if user := authenticate(request, **serializer.validated_data):
            login(request, user)
            return Response(UserSerializer(user).data)
        else:
            raise AuthenticationFailed()

    @extend_schema(responses={200: UserSerializer})
    @action(detail=False, methods=["get"], permission_classes=[IsAuthenticated])
    def me(self, request: Request) -> Response:
        return Response(UserSerializer(request.user).data)

    @extend_schema(operation_id="user_logout", responses={204: None})
    @action(
        detail=False,
        methods=["post"],
        permission_classes=[IsAuthenticated],
    )
    def logout(self, request: Request) -> Response:
        logout(request)
        return Response(status=status.HTTP_204_NO_CONTENT)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-01-09 20:53
from __future__ import unicode_literals

from datetime import timedelta
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


def set_due_dates(apps, schema_editor):
    Event = apps.get_model('leprikon', 'Event')
    Event.objects.update(
        due_from=models.F('start_date') - timedelta(180),
        due_date=models.F('start_date'),
    )

    SchoolYearPeriod = apps.get_model('leprikon', 'SchoolYearPeriod')
    SchoolYearPeriod.objects.update(
        due_from=models.F('start') - timedelta(30),
        due_date=models.F('start') + timedelta(30),
    )


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0039_editable_texts'),
    ]

    operations = [
        migrations.AddField(
            model_name='event',
            name='due_date',
            field=models.DateField(default=django.utils.timezone.now, verbose_name='due date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='event',
            name='due_from',
            field=models.DateField(default=django.utils.timezone.now, verbose_name='due from'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='schoolyearperiod',
            name='due_date',
            field=models.DateField(default=django.utils.timezone.now, verbose_name='due date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='schoolyearperiod',
            name='due_from',
            field=models.DateField(default=django.utils.timezone.now, verbose_name='due from'),
            preserve_default=False,
        ),
        migrations.RunPython(set_due_dates),
    ]



# Generated by Django 3.2.20 on 2023-09-06 14:30

from django.db import migrations, models


def copy_min_due_date_days(apps, schema_editor):
    Event = apps.get_model("leprikon", "Event")
    for event in Event.objects.iterator():
        event.min_due_date_days = event.min_due_date_days_old
        event.save()


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0080_delete_courseregistrationhistory"),
    ]

    operations = [
        migrations.RenameField(
            model_name="subject",
            old_name="min_due_date_days",
            new_name="min_due_date_days_old",
        ),
        migrations.AddField(
            model_name="event",
            name="min_due_date_days",
            field=models.PositiveIntegerField(default=3, verbose_name="minimal number of days to due date"),
        ),
        migrations.AddField(
            model_name="orderable",
            name="min_due_date_days",
            field=models.PositiveIntegerField(default=3, verbose_name="minimal number of days to due date"),
        ),
        migrations.AddField(
            model_name="schoolyearperiod",
            name="min_due_date_days",
            field=models.PositiveIntegerField(default=14, verbose_name="minimal number of days to due date"),
        ),
        migrations.RunPython(copy_min_due_date_days),
        migrations.RemoveField(
            model_name="subject",
            name="min_due_date_days_old",
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-06-19 22:05
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0069_question_slug"),
    ]

    operations = [
        migrations.AlterField(
            model_name="question",
            name="slug",
            field=models.SlugField(unique=True, verbose_name="unique identifier"),
        ),
        migrations.AddField(
            model_name="leprikonsite",
            name="decision_print_setup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="default decision print setup",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="decision_print_setup",
            field=models.ForeignKey(
                blank=True,
                help_text="Only use to set value specific for this subject.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="registration print setup",
            ),
        ),
        migrations.AddField(
            model_name="subjecttype",
            name="decision_print_setup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="decision print setup",
            ),
        ),
    ]
